#!/usr/bin/env python
"""
utils.py
"""
import os, warnings, numpy as np, pandas as pd 
from glob import glob 
from typing import List
from itertools import accumulate, chain, repeat
from .constants import FRAME, TRACK, TRACK_LENGTH, PY, PX

######################
## TRACKS UTILITIES ##
######################


def track_length(tracks: pd.DataFrame) -> pd.DataFrame:
    """
    Add a new column to a trajectory dataframe with the trajectory
    length in frames.

    args
    ----
        tracks      :   pandas.DataFrame. Must have the column 
                        *TRACK*.

    returns
    -------
        pandas.DataFrame, with the "track_length" column. Overwritten
            if it already exists.

    """
    if TRACK_LENGTH in tracks.columns:
        tracks = tracks.drop(TRACK_LENGTH, axis=1)
    return tracks.join(tracks.groupby(TRACK).size().rename(TRACK_LENGTH), on=TRACK)


def assign_index_in_track(tracks):
    """
    Given a set of trajectories, determine the index of each localization in the
    context of its respective trajectory.

    args
    ----
        tracks      :   pandas.DataFrame, containing the "trajectory" and "frame"
                        columns

    returns
    -------
        pandas.DataFrame, the same dataframe with a new column, "index_in_track"

    """
    tracks = tracks.sort_values(by=[TRACK, FRAME])
    tracks["one"] =  1
    tracks["index_in_track"] = tracks.groupby(TRACK)["one"].cumsum() - 1
    tracks = tracks.drop("one", axis=1)
    return tracks 


def concat_tracks(*tracks):
    """
    Join some trajectory dataframes together into a larger dataframe,
    while preserving uniqe trajectory indices.

    args
    ----
        tracks      :   pandas.DataFrame with the "trajectory" column

    returns
    -------
        pandas.DataFrame, the concatenated trajectories

    """
    n = len(tracks)

    # Sort the tracks dataframes by their size. The only important thing
    # here is that if at least one of the tracks dataframes is nonempty,
    # we need to put that one first.
    df_lens = [len(t) for t in tracks]
    try:
        tracks = [t for _, t in sorted(zip(df_lens, tracks))][::-1]
    except ValueError:
        pass

    # Iteratively concatenate each dataframe to the first while 
    # incrementing the trajectory index as necessary
    out = tracks[0].assign(dataframe_index=0)
    c_idx = out[TRACK].max() + 1

    for t in range(1, n):

        # Get the next set of trajectories and keep track of the origin
        # dataframe
        new = tracks[t].assign(dataframe_index=t)

        # Ignore negative trajectory indices (facilitating a user filter)
        new.loc[new["trajectory"]>=0, "trajectory"] += c_idx 

        # Increment the total number of trajectories
        c_idx = new["trajectory"].max() + 1

        # Concatenate
        out = pd.concat([out, new], ignore_index=True, sort=False)

    return out


#############################
## TRACK LOADING UTILITIES ##
#############################


def load_tracks(*csv_paths, out_csv=None, start_frame=0,
    drop_singlets=False, suffix=".csv"):
    """
    Given a set of trajectories stored as CSVs, concatenate all
    of them, storing the paths to the original CSVs in the resulting
    dataframe, and optionally save the result to another CSV.
    If passed a directory instead of a set of CSV paths, find all 
    the CSVs in that directory that end with *suffix*, load the 
    trajectories, and concatenate them.

    args
    ----
        csv_paths       :   list of str, a set of trajectory CSVs.
                            Each must contain the "y", "x", "trajectory",
                            and "frame" columns
        out_csv         :   str, path to save to 
        start_frame     :   int, exclude any trajectories that begin before
                            this frame
        drop_singlets   :   bool, drop singlet localizations before
                            concatenating
        suffix          ;   str, suffix of CSVs if passing a directory

    returns
    -------
        pandas.DataFrame, the concatenated result

    """
    n = len(csv_paths)
    if n == 0:
        warnings.warn("no paths passed")
        return pd.DataFrame([], columns=["trajectory", "frame", "y", "x"], dtype=object)

    # If passed a directory instead of a set of file paths, just load all 
    # the CSVs from that directory
    if os.path.isdir(csv_paths[0]):
        return load_tracks_dir(csv_paths[0], start_frame=start_frame,
            drop_singlets=drop_singlets, suffix=suffix)

    if start_frame is None:
        start_frame = 0

    def drop_before_start_frame(tracks, start_frame):
        """
        Drop all trajectories that start before a specific frame.
        """
        if tracks.empty or (start_frame is None) or (start_frame <= tracks["frame"].min()):
            return tracks 

        tracks = tracks.join(
            (tracks.groupby("trajectory")["frame"].first() >= start_frame).rename("_take"),
            on="trajectory"
        )

        tracks = tracks[tracks["_take"]]
        tracks = tracks.drop("_take", axis=1)

        return tracks

    def drop_singlets_dataframe(tracks):
        """
        Drop all singlets and unassigned localizations from a 
        pandas.DataFrame with trajectory information.
        """
        if tracks.empty:
            return tracks 

        tracks = track_length(tracks)
        tracks = tracks[np.logical_and(tracks["track_length"]>1,
            tracks["trajectory"]>=0)]

        return tracks 

    def loader(path):
        tracks = pd.read_csv(path)

        if drop_singlets:
            tracks = drop_singlets_dataframe(tracks)

        tracks = drop_before_start_frame(tracks, start_frame)

        return tracks 

    # Load the trajectories into memory
    tracks = []
    for path in csv_paths:
        tracks.append(loader(path))

    # Concatenate 
    tracks = concat_tracks(*tracks)

    # Map the original path back to each file
    for i, path in enumerate(csv_paths):
        tracks.loc[tracks["dataframe_index"]==i, "source_file"] = \
            os.path.abspath(path)

    # Optionally save concatenated trajectories to a new CSV
    if not out_csv is None:
        tracks.to_csv(out_csv, index=False)

    return tracks 


def load_tracks_dir(dirname, suffix=".csv", start_frame=0,
    drop_singlets=False):
    """
    Load all of the trajectory CSVs in a target directory
    into a single pandas.DataFrame.

    args
    ----
        dirname         :   str, directory with the track CSVs
        suffix          :   str, extension for the track CSVs
        start_frame     :   int, exclude all tracks before this
                            frame
        drop_singlets   :   bool, don't include single-point 
                            trajectories
    returns
    -------
        pandas.DataFrame with an extra column, "origin_file",
            with the path to the CSV from which these 
            trajectories were taken

    """
    # Find target files
    if os.path.isdir(dirname):
        target_csvs = glob(os.path.join(dirname, "*{}".format(suffix)))
        if len(target_csvs) == 0:
            raise IOError("Could not find trajectory CSVs in directory {}".format(dirname))
    elif os.path.isfile(dirname):
        target_csvs = [dirname]

    # Concatenate trajectories
    tracks = [pd.read_csv(j) for j in target_csvs]
    tracks = concat_tracks(*tracks)

    # Exclude points before the start frame
    if isinstance(start_frame, int) and \
        (start_frame > tracks["frame"].min()) and \
        (not tracks.empty):

        tracks = tracks.join(
            (tracks.groupby("trajectory")["frame"].first() >= start_frame).rename("_take"),
            on="trajectory"
        )
        tracks = tracks[tracks["_take"]] 
        tracks = tracks.drop("_take", axis=1)

    # Exclude trajectories that are too short
    tracks = track_length(tracks)
    if drop_singlets:
        tracks = tracks[tracks["track_length"] > 1]

    return tracks 


####################
## JUMP COMPUTERS ##
####################


def tracks_to_jumps(tracks, n_frames=1, start_frame=None, pixel_size_um=0.16,
    pos_cols=["y", "x"]):
    """
    Convert trajectories in pandas.DataFrame format to an internal "jumps"
    format, specified in the *returns* section of this docstring.

    args
    ----
        tracks          :   pandas.DataFrame
        n_frames        :   int, the number of frames over which to compute
                            the jump. For instance, if n_frames = 1, then only
                            compute jumps between consecutive frames
        start_frame     :   int, disregard jumps before this frame
        pixel_size_um   :   float, size of pixels in microns
        pos_cols        :   list of str, the columns with the spatial 
                            coordinates of each point in pixels

    returns
    -------
        *jumps*, a 2D ndarray of shape (n_jumps, 6+). Each row corresponds
            to a single jump from the dataset. 
        The columns of *vecs* have the following meaning:
            jumps[:,0] -> length of the origin trajectory in frames
            jumps[:,1] -> index of the origin trajectory in *tracks*
            jumps[:,2] -> frame corresponding to the first point in 
                            the jump
            jumps[:,3] -> sum of squared jumps across all spatial
                            dimensions in squared microns
            jumps[:,4:] -> jumps in each Euclidean dimension in microns

    """
    def bail():
        return np.zeros((0, 6), dtype=np.float64)

    # If passed an empty dataframe, bail
    if tracks.empty: return bail()

    # Do not modify the original dataframe
    tracks = tracks.copy()

    # Calculate the original trajectory length and exclude
    # singlets and negative trajectory indices
    tracks = track_length(tracks)
    tracks = tracks[np.logical_and(
        tracks["trajectory"] >= 0,
        tracks["track_length"] > 1
    )]

    # Only consider trajectories after some start frame
    if not start_frame is None:
        tracks = tracks[tracks["frame"] >= start_frame]

    # If no trajectories remain, bail
    if tracks.empty: return bail()

    # Convert from pixels to um
    tracks[pos_cols] *= pixel_size_um 

    # Work with an ndarray, for speed
    tracks = tracks.sort_values(by=["trajectory", "frame"])
    T = np.asarray(tracks[["track_length", "trajectory", "frame", pos_cols[0]] + pos_cols])

    # Allowing for gaps, consider every possible comparison that 
    # leads to the correct frame interval
    target_jumps = []
    for j in range(1, n_frames+1):
        
        # Compute jumps
        jumps = T[j:,:] - T[:-j,:]

        # Only consider vectors between points originating 
        # from the same trajectory and from the target frame
        # interval
        same_track = jumps[:,1] == 0
        target_interval = jumps[:,2] == n_frames 
        take = np.logical_and(same_track, target_interval)

        # Map the corresponding track lengths, track indices,
        # and frame indices back to each jump
        jumps[:,:3] = T[:-j,:3]
        jumps = jumps[take, :]

        # Calculate the corresponding 2D squared jump and accumulate
        if jumps.shape[0] > 0:
            jumps[:,3] = (jumps[:,4:]**2).sum(axis=1)
            target_jumps.append(jumps)

    # Concatenate
    if len(target_jumps) > 0:
        return np.concatenate(target_jumps, axis=0)
    else:
        return bail()


def sum_squared_jumps(jumps, max_jumps_per_track=None, pos_cols=["y", "x"]):
    """
    For each trajectory in a dataset, calculate the sum of its squared
    jumps across all spatial dimensions.
    args
    ----
        jumps           :   2D ndarray, all jumps in the dataset as 
                            calculated by *tracks_to_jumps*
        max_jumps_per_track :   int, the maximum number of jumps 
                            to consider from any single trajectory
    returns
    -------
        pandas.DataFrame. Each row corresponds to a trajectory, with
            the following columns:
            "sum_sq_jump": the summed squared jumps of that trajectory
                           in microns
            "trajectory" : the index of the origin trajectory
            "frame"      : the first frame of the first jumps in the 
                           origin trajectory
            "n_jumps"    : the number of jumps used in *sum_sq_jump*
    """
    out_cols = ["sum_sq_jump", "trajectory", "frame", "n_jumps"]

    # If there are no jumps in this set of trajectories, bail
    if jumps.shape[0] == 0:
        return pd.DataFrame(index=[], columns=out_cols, dtype=object)

    # Format as a dataframe, indexed by jump
    cols = ["track_length", "trajectory", "frame", "sq_jump"] + list(pos_cols)
    jumps = pd.DataFrame(jumps, columns=cols)
    n_tracks = jumps["trajectory"].nunique()

    # Limit the number of jumps to consider per trajectory, if desired
    if not max_jumps_per_track is None:
        jumps = assign_index_in_track(jumps)
        tracks = jumps[jumps["index_in_track"] <= max_jumps_per_track]

    # Output dataframe, indexed by trajectory
    sum_jumps = pd.DataFrame(index=np.arange(n_tracks), columns=out_cols, dtype=object)

    # Calculate the sum of squared jumps for each trajectory
    sum_jumps["sum_sq_jump"] = np.asarray(jumps.groupby("trajectory")["sq_jump"].sum())

    # Calculate the number of jumps in each trajectory
    sum_jumps["n_jumps"] = np.asarray(jumps.groupby("trajectory").size())

    # Map back the indices of the origin trajectories
    sum_jumps["trajectory"] = np.asarray(jumps.groupby("trajectory").apply(lambda i: i.name)).astype(np.int64)

    # Map back the frame indices
    sum_jumps["frame"] = np.asarray(jumps.groupby("trajectory")["frame"].first()).astype(np.int64)

    return sum_jumps


def split_jumps(jumps, splitsize=8):
    """
    Split a set of long trajectories into shorter trajectories.

    Example 1
    ---------
        If we have a trajectory of 6 jumps and 
        splitsize = 3, then we split this trajectory into two 
        trajectories of 3 jumps, comprising the first and second halves
        of the original trajectory.

    Example 2
    ---------
        If we have a trajectory of 10 jumps and splitsize = 4,
        then we split this trajectory into 3 trajectories. The 
        first two are 4 jumps each, and the third is the last 2
        jumps of the original trajectory.

    args
    ----
        jumps           :   2D ndarray, a set of trajectory-indexed
                            jumps; output of *tracks_to_jumps*
        splitsize       :   int, the maximum size of a trajectory 
                            after splitting

    returns
    -------
        1D ndarray of shape (n_tracks), the indices of the 
            new trajectories. These start from 0 and go to the 
            highest new trajectory index; numerically they have 
            no relation to the original trajectory indices.
            
    """
    # If passed empty input, return empty output
    if jumps.shape[0] == 0:
        return np.zeros(0, dtype=np.int64)

    # The original set of trajectory indices
    orig_indices = jumps[:,1].astype(np.int64)

    # The set of modified trajectory indices
    new_indices = np.zeros(orig_indices.shape[0], dtype=np.int64)

    # The current (new) trajectory index 
    c = 0

    # The length of the current trajectory in # of jumps
    L = 0

    # Iterate through the original set of trajectory indices
    prev_index = orig_indices[0]
    for i, index in enumerate(orig_indices):

        # Extend the existing trajectory
        L += 1

        # We're in the same original trajectory
        if index == prev_index:

            # Haven't exceeded the split trajectory size limit
            if L < splitsize:
                new_indices[i] = c 

            # Break into a new trajectory
            else:
                L = 0
                c += 1
                new_indices[i] = c

        # We've passed into a different original trajectory
        else:
            prev_index = index 
            L = 0
            c += 1
            new_indices[i] = c 

    return new_indices

#####################
## OTHER UTILITIES ##
#####################

def normalize_2d(arr: np.ndarray, axis: int):
    """ Normalize a 2D array over one of its axes.

    args
    ----
        arr         :   2D numpy.ndarray
        axis        :   int, axis to normalize over

    returns
    -------
        *arr* such as that arr.sum(axis=axis) == 1.0
    """
    if len(arr.shape) != 2:
        raise ValueError(f"arr has shape {arr.shape}; expected 2D")

    if axis == 1:
        S = arr.sum(axis=1)
        nonzero = S > 0
        arr[nonzero,:] = (arr[nonzero,:].T / S[nonzero]).T
    else:
        S = arr.sum(axis=0)
        nonzero = S > 0
        arr[:,nonzero] = arr[:,nonzero] / S[nonzero]
    return arr

def cartesian_product(*arrays: np.ndarray):
    """ Cartesian product of multiple 1D numpy.ndarrays.

    Source: https://stackoverflow.com/a/45378609

    args
    ----
        arrays  :   1D numpy.ndarray

    returns
    -------

    """
    la = len(arrays)
    L = *map(len, arrays), la
    dtype = np.result_type(*arrays)
    arr = np.empty(L, dtype=dtype)
    arrs = *accumulate(chain((arr,), repeat(0, la-1)), np.ndarray.__getitem__),
    idx = slice(None), *repeat(None, la-1)
    for i in range(la-1, 0, -1):
        arrs[i][..., i] = arrays[i][idx[:la-i]]
        arrs[i-1][1:] = arrs[i]
    arr[..., 0] = arrays[0][idx]
    return arr.reshape(-1, la)



"""
PLOTS TO MAKE:
    - Likelihood-specific functions
        - summarize RBME output on a single set of trajectories
    - Dataset-level functions
        - Show marginal likelihoods by experimental condition
        - Show marginal posterior occupations by experimental condition (heat map)
        - Show marginal posterior occupations by experimental condition (line plot)
            (perhaps enable a "hue" that can overlay sub-conditions, such as 
            molecule concentration)
"""
import os, sys, warnings, numpy as np, pandas as pd, matplotlib, \
    matplotlib.pyplot as plt, matplotlib.gridspec as grd, seaborn as sns
from typing import Tuple, List
from mpl_toolkits.axes_grid1 import make_axes_locatable
from matplotlib.cm import get_cmap
from scipy.ndimage import gaussian_filter

from .constants import TRACK, FRAME, PY, PX
from .utils import normalize_2d

matplotlib.rcParams['font.family'] = 'sans-serif'
matplotlib.rcParams['font.sans-serif'] = 'Helvetica'

########################
## PLOTTING UTILITIES ##
########################

def kill_ticks(axes: matplotlib.axes.Axes, spines: bool=False, grid: bool=False):
    """
    Remove the ticks and/or splines and/or grid from a
    matplotlib Axes.
    args
    ----
        axes        :   matplotlib.axes.Axes
        spines      :   bool, also remove spines
        grid        :   boo, also remove the grid
    """
    axes.set_xticks([])
    axes.set_yticks([])
    if spines:
        for s in ['top', 'bottom', 'left', 'right']:
            axes.spines[s].set_visible(False)
    if grid:
        axes.grid(False)

def save_png(out_png: str, dpi: int=600):
    """
    Save a matplotlib.figure.Figure to a PNG.
    args
    ----
        out_png         :   str, out path
        dpi             :   int, resolution
    """
    plt.tight_layout()
    plt.savefig(out_png, dpi=dpi)
    plt.close()

def nanpercentile(arr: np.ndarray, perc: float) -> float:
    """ Thin wrapper on numpy.nanpercentile that returns 1.0 when the
    input is all NaN or empty. """
    if arr.size == 0 or np.isnan(arr).all():
        return 1.0
    else:
        return np.nanpercentile(arr, perc)

def add_log_scale_imshow(axes: matplotlib.axes.Axes, diff_coefs: np.ndarray,
    fontsize: int=None, side: str="x"):
    """
    Add a log axis to a plot produced by matplotlib.axes.Axes.imshow.
    This is specifically used to show the log-values of the diffusion 
    coefficient corresponding to each (linear) x-axis point in some
    of the likelihood plots.
    args
    ----
        axes        :   matplotlib.axes.Axes
        diff_coefs  :   1D ndarray
        fontsize    :   int
        side        :   str, "x" or "y"
    returns
    -------
        None; modifies *axes* directly
    """
    diff_coefs = np.asarray(diff_coefs)
    K = diff_coefs.shape[0]
    d_min = min(diff_coefs)
    d_max = max(diff_coefs)

    # Linear range of the axes
    if side == "x":
        lim = axes.get_xlim()
    elif side == "y":
        lim = axes.get_ylim()

    # xlim = axes.get_xlim()
    lin_span = lim[1] - lim[0]

    # Determine the number of log-10 units (corresponding
    # to major ticks)
    log_diff_coefs = np.log10(diff_coefs)
    first_major_tick = int(log_diff_coefs[0])
    major_tick_values = [first_major_tick]
    c = first_major_tick
    while log_diff_coefs.max() > c:
        c += 1
        major_tick_values.append(c)
    n_major_ticks = len(major_tick_values)

    # Convert between the linear and log scales
    log_span = log_diff_coefs[-1] - log_diff_coefs[0]
    m = lin_span / log_span 
    b = lim[0] - m * log_diff_coefs[0]
    def convert_log_to_lin_coord(log_coord):
        return m * log_coord + b

    # Choose the location of the major ticks
    major_tick_locs = [convert_log_to_lin_coord(coord) \
        for coord in major_tick_values]

    # Major tick labels
    major_tick_labels = ["$10^{%d}$" % int(j) for j in major_tick_values]

    # Minor ticks 
    minor_tick_decile = np.log10(np.arange(1, 11))
    minor_tick_values = []
    for i in range(int(major_tick_values[0])-1, int(major_tick_values[-1])+2):
        minor_tick_values += list(minor_tick_decile + i)
    minor_tick_locs = [convert_log_to_lin_coord(v) for v in minor_tick_values]
    minor_tick_locs = [i for i in minor_tick_locs if ((i >= lim[0]) and (i <= lim[1]))]

    # Set the ticks
    if side == "x":
        axes.set_xticks(major_tick_locs, minor=False)
        axes.set_xticklabels(major_tick_labels, fontsize=fontsize)
        axes.set_xticks(minor_tick_locs, minor=True)
    elif side == "y":
        axes.set_yticks(major_tick_locs, minor=False)
        axes.set_yticklabels(major_tick_labels, fontsize=fontsize)
        axes.set_yticks(minor_tick_locs, minor=True)

###############################
## LIKELIHOOD-SPECIFIC PLOTS ##
###############################

def rbme_posterior_plot(
    out_png: str,
    marginal_likelihood: np.ndarray,
    posterior_occs: np.ndarray,
    diff_coefs: np.ndarray,
    loc_errors: np.ndarray,
    truncate_y_axis: bool=True,
    suptitle: str=None,
):
    """ Visualize the posterior distribution of a state array defined
    with the RBME likelihood function. This likelihood function takes two 
    parameters: diffusion coefficient and localization error. The plot has
    three vertically aligned axes that show the marginal likelihood, mean
    posterior occupations, and marginal posterior occupations respectively.

    args
    ----
        out_png                 :   file to save plot to
        marginal_likelihood     :   2D numpy.ndarray of shape (n_diff_coefs, n_loc_errors)
        posterior_occs          :   2D numpy.ndarray of shape (n_diff_coefs, n_loc_errors)
        diff_coefs              :   1D numpy.ndarray of shape (n_diff_coefs,)
        loc_errors              :   1D numpy.ndarray of shape (n_loc_errors,)
        truncate_y_axis         :   cut off the immobile fraction peak for visibility
        suptitle                :   str, main figure title

    """
    # Catch unsanitary input
    if (not len(diff_coefs.shape) == 1) or (not len(loc_errors.shape) == 1):
        raise ValueError("diff_coefs and loc_errors must be 1-dimensional")
    if not all(map(lambda p: len(p.shape)==2, [marginal_likelihood, posterior_occs])):
        raise ValueError("marginal_likelihood and posterior_occs must be 2-dimensional")
    if not marginal_likelihood.shape == (diff_coefs.shape[0], loc_errors.shape[0]):
        raise ValueError(f"mismatch in shapes of marginal_likelihood " \
            "({marginal_likelihood.shape}), diff_coefs ({diff_coefs.shape}), " \
            "and/or loc_errors ({loc_errors.shape})")
    if not posterior_occs.shape == (diff_coefs.shape[0], loc_errors.shape[0]):
        raise ValueError(f"mismatch in shapes of posterior_occs " \
            "({posterior_occs.shape}), diff_coefs ({diff_coefs.shape}), " \
            "and/or loc_errors ({loc_errors.shape})")

    # Marginalize posterior occupations on localization error
    posterior_occs_marg = posterior_occs.sum(axis=1)
    posterior_occs_marg /= posterior_occs_marg.sum()

    # Plot layout
    fig, ax = plt.subplots(figsize=(6, 5))
    fontsize = 10
    gs = grd.GridSpec(3, 1, height_ratios=(2, 2, 1), width_ratios=None, hspace=0.75)
    ax0 = plt.subplot(gs[0])
    ax1 = plt.subplot(gs[1])
    ax2 = plt.subplot(gs[2])
    ax = [ax0, ax1, ax2]

    # Color map scaling
    vmax0 = np.percentile(marginal_likelihood, 99.5)
    vmax1 = np.percentile(posterior_occs, 99.5)

    p0 = ax[0].imshow(marginal_likelihood.T, cmap="viridis", origin="lower",
        aspect="auto", vmin=0, vmax=vmax0)
    p1 = ax[1].imshow(posterior_occs.T, cmap="viridis", origin="lower",
        aspect="auto", vmin=0, vmax=vmax1)

    # Add log scales for the x-axis
    add_log_scale_imshow(ax[0], diff_coefs, side="x")
    add_log_scale_imshow(ax[1], diff_coefs, side="x")

    # y-ticks for localization error
    n_yticks = 5
    space = loc_errors.shape[0] // n_yticks
    if space > loc_errors.shape[0]:
        space = 1
    try:
        yticks = np.arange(loc_errors.shape[0])[::space]
        yticklabels = ['%.3f' % i for i in loc_errors[::space]]
    except:
        space = 1
        yticks = np.arange(loc_errors.shape[0])[::space]
        yticklabels = ['%.3f' % i for i in loc_errors[::space]]
    for j in range(2):
        ax[j].set_yticks(yticks)
        ax[j].set_yticklabels(yticklabels, fontsize=fontsize)
        ax[j].set_ylabel("Localization error ($\mu$m)", fontsize=fontsize)

    # Show the posterior mean marginalized on localization error
    ax[2].plot(diff_coefs, posterior_occs_marg, color='k')
    ax[2].set_xscale("log")
    ax[2].set_ylabel("Marginal posterior\noccupation", fontsize=fontsize)
    ax[2].set_xlim((np.min(diff_coefs), np.max(diff_coefs)))
    if truncate_y_axis:
        ax[2].set_ylim((0, posterior_occs_marg[diff_coefs>0.05].max()*2.0))
    else:
        ax[2].set_ylim((0, max(posterior_occs_marg)))
    ax[2].set_xlabel("Diffusion coefficient ($\mu$m$^{2}$ s$^{-1}$)", fontsize=fontsize)

    # Subplot titles
    ax[0].set_title("Naive occupation", fontsize=fontsize)
    ax[1].set_title("Posterior occupation", fontsize=fontsize)

    # Main figure title
    if suptitle is not None:
        fig.suptitle(suptitle, fontsize=fontsize)

    # Save, ignoring complaints from a call to matplotlib.pyplot.tight_layout()
    with warnings.catch_warnings(record=False) as w:
        warnings.simplefilter("ignore")
        save_png(out_png, dpi=600)

def rbm_posterior_plot(
    out_png: str,
    marginal_likelihood: np.ndarray,
    posterior_occs: np.ndarray,
    diff_coefs: np.ndarray,
    truncate_y_axis: bool=True,
    fontsize: int=8,
    suptitle: str=None,
):
    """ Plot a 1-dimensional posterior distribution over a grid of 
    logarithmically-spaced diffusion coefficients.

    args
    ----
        out_png                 :   file to save plot to
        marginal_likelihood     :   1D numpy.ndarray of shape
                                    (n_diff_coefs), likelihood function
                                    marginalized over trajectory-state
                                    assignments
        posterior_occs          :   1D numpy.ndarray of shape 
                                    (n_diff_coefs), posterior mean state
                                    occupations
        diff_coefs              :   1D numpy.ndarray of shape 
                                    (n_diff_coefs), diffusion coefficient
                                    for each state in µm2/sec
        truncate_y_axis         :   bool, cut off the immobile fraction
                                    peak if it's too large
        fontsize                :   int
        suptitle                :   str, a title for the plot
    """
    if not all([
        posterior_occs.shape[0] == diff_coefs.shape[0],
        marginal_likelihood.shape[0] == diff_coefs.shape[0]
    ]):
        raise ValueError(f"shape mismatch: {posterior_occs.shape}, " \
            "{marginal_likelihood.shape}, {diff_coefs.shape}")
    posterior_occs = posterior_occs.copy() / np.nansum(posterior_occs)
    fig, ax = plt.subplots(2, 1, figsize=(5, 2.5), sharex=True)
    ax[0].plot(diff_coefs, marginal_likelihood, color='k')
    ax[1].plot(diff_coefs, posterior_occs, color='k')
    ax[1].set_xscale("log")
    ax[1].set_xlabel("Diffusion coefficient ($\mu$m$^{2}$ s$^{-1}$)", fontsize=fontsize)
    ax[0].set_ylabel("Naive\noccupation", fontsize=fontsize)
    ax[1].set_ylabel("Posterior\noccupation", fontsize=fontsize)
    for j in range(2):
        ax[j].set_ylim((0, ax[j].get_ylim()[1]))
        ax[j].yaxis.set_tick_params(labelsize=fontsize)
        ax[j].xaxis.set_tick_params(labelsize=fontsize)
    if suptitle is not None:
        fig.suptitle(suptitle, fontsize=fontsize)
    if truncate_y_axis:
        ax[1].set_ylim((0, np.nanmax(posterior_occs[diff_coefs>0.05])*2.0))
    plt.tight_layout()
    plt.savefig(out_png, dpi=600)
    plt.close()

#######################################
## TRAJECTORY-STATE ASSIGNMENT PLOTS ##
#######################################

def marginal_assignment_probability_plot(
    out_png: str, 
    assignment_probabilities: np.ndarray, 
    assignment_likelihoods: np.ndarray,
    diff_coefs: np.ndarray,
    subplot_extent: Tuple[int]=(0, 6, 0, 1.5),
    fontsize: int=10,
    sort_by_mean_diff_coef: bool=False,
):
    """ Visualize the trajectory-state assignment likelihoods and
    probabilities. The upper subplot shows likelihoods and the lower
    subplot shows probabilities (for instance, the posterior assignment
    probability from a state array run).

    In each subplot, trajectories correspond to columns of a heat map while 
    diffusive states correspond to rows.

    args
    ----
        out_png                 :   output file
        assignment_probabilities:   2D numpy.ndarray of shape (n_diff_coefs,
                                    n_tracks)
        assignment_likelihoods  :   2D numpy.ndarray of shape (n_diff_coefs,
                                    n_tracks)
        diff_coefs              :   1D numpy.ndarray of shape (n_diff_coefs,)
                                    diffusion coefficient in µm2/sec
        subplot_extent          :   4-tuple of int, shape of subplots
        fontsize                :   int
        sort_by_mean_diff_coef  :   bool, plot each trajectory in order of 
                                    increasing mean diffusion coefficient

    """
    figsize = (subplot_extent[1], subplot_extent[3]*2)
    fig, ax = plt.subplots(2, 1, figsize=figsize, sharex=True, sharey=True)

    # Normalize
    if assignment_probabilities.size > 0:
        assignment_probabilities = assignment_probabilities.copy() / np.nansum(assignment_probabilities, axis=0)

    # Sort values by mean posterior diffusion coefficient
    if sort_by_mean_diff_coef:
        indices = np.argsort((assignment_probabilities.T * diff_coefs).sum(axis=1))
        assignment_probabilities = assignment_probabilities[:,indices]
        assignment_likelihoods = assignment_likelihoods[:,indices]

    # Main plot
    f0 = ax[0].imshow(assignment_likelihoods, origin="lower", cmap='viridis', 
        extent=subplot_extent, vmax=nanpercentile(assignment_likelihoods, 99))
    f1 = ax[1].imshow(assignment_probabilities, origin="lower", cmap='viridis',
        extent=subplot_extent, vmin=0, vmax=nanpercentile(assignment_probabilities, 99))

    # Color scale
    cbar = plt.colorbar(f0, ax=ax[0], shrink=0.75)
    cbar.set_label("Naive probability", fontsize=fontsize)
    cbar.ax.tick_params(labelsize=fontsize)
    cbar = plt.colorbar(f1, ax=ax[1], shrink=0.75)
    cbar.set_label("Posterior probability", fontsize=fontsize)
    cbar.ax.tick_params(labelsize=fontsize)

    # Ticks and axis labels
    ax[1].set_xticks([])
    ax[1].set_xlabel("Trajectory", fontsize=fontsize)
    add_log_scale_imshow(ax[1], diff_coefs, fontsize=None, side='y')
    for j in range(2):
        ax[j].set_ylabel("Diffusion\ncoefficient\n($\mu$m$^{2}$ s$^{-1}$)",
            fontsize=fontsize)

    # Save
    plt.tight_layout()
    plt.savefig(out_png, dpi=600)
    plt.close()

def temporal_assignment_probability_plot(
    out_png: str,
    tracks: pd.DataFrame,
    assignment_probabilities: np.ndarray,
    assignment_likelihoods: np.ndarray,
    diff_coefs: np.ndarray,
    subplot_extent: Tuple[int]=(0, 6, 0, 1.5),
    fontsize: int=10,
    tick_fontsize: int=8,
    frame_block_size: int=10,
    normalize: bool=False,
    suptitle: str=None,
):
    """ Plot the trajectory-state assignment probabilities vs. frame index.
    Useful for visualizing density-dependent effects on the assignment 
    probabilities.

    args
    ----
        out_png                 :   output filename
        tracks                  :   pandas.DataFrame, detections
        assignment_probabilities:   2D numpy.ndarray of shape (n_diff_coefs, n_tracks)
        assignment_likelihoods  :   2D numpy.ndarray of shape (n_diff_coefs, n_tracks)
        diff_coefs              :   1D numpy.ndarray of shape (n_diff_coefs,)
        subplot_extent          :   4-tuple of int
        fontsize                :   int
        frame_block_size        :   bin size for aggregating assignment
                                    probabilities/likelihoods
        normalize               :   normalize within each frame block
        suptitle                :   main figure title

    """
    if not all([
        len(assignment_probabilities.shape) == 2,
        len(assignment_likelihoods.shape) == 2,
        len(diff_coefs.shape) == 1,
        assignment_probabilities.shape[0] == diff_coefs.shape[0],
        assignment_likelihoods.shape[0] == diff_coefs.shape[0],
    ]):
        raise ValueError("incompatible shapes for temporal " \
            "assignment probability plot")

    # Number of distinct diffusion coefficients, frames, and frame blocks
    n_diff_coefs = diff_coefs.shape[0]
    n_frames = np.nanmax(tracks[FRAME]) + 1 if len(tracks) > 0 else 1
    n_blocks = (n_frames + frame_block_size - 1) // frame_block_size

    # Index of each trajectory
    track_indices = np.arange(assignment_probabilities.shape[1])

    # Only include track indices represented in this set of trajectories
    if not np.isin(track_indices, tracks[TRACK].unique()).all():
        raise ValueError(f"not all trajectory indices are present in this set of detections")

    # Start frame and stop frame for each trajectory
    start_frames = np.asarray(tracks.groupby(TRACK)[FRAME].first().loc[track_indices])
    stop_frames = np.asarray(tracks.groupby(TRACK)[FRAME].last().loc[track_indices])

    # Posterior probability and likelihood aggregated by frame
    P = np.full((n_diff_coefs, n_blocks), np.nan, dtype=np.float64)
    L = np.full((n_diff_coefs, n_blocks), np.nan, dtype=np.float64)

    for b in range(n_blocks):
        in_block = np.logical_and(
            start_frames<(b+1)*frame_block_size,
            stop_frames>=b*frame_block_size
        )
        if in_block.any():
            P[:,b] = assignment_probabilities[:,in_block].sum(axis=1)
            L[:,b] = assignment_likelihoods[:,in_block].sum(axis=1)

    # Detections per frame
    tracks["frame_block"] = tracks[FRAME] // frame_block_size
    detections_per_frame = np.asarray(
        pd.Series(np.arange(n_blocks)).map(
            tracks.groupby("frame_block").size()
        )
    ) / frame_block_size
    tracks = tracks.drop("frame_block", axis=1)

    # Normalize for each block
    if normalize:
        def norm(x):
            s = x.sum(axis=0)
            nonzero = s > 0
            x[:,nonzero] = x[:,nonzero] / s[nonzero]
            return x 
        P = norm(P)
        L = norm(L)

    # Plot layout
    subplot_extent = (0, 6, 0, 1.5)
    figsize = (subplot_extent[1], subplot_extent[3]*3)
    fig, ax = plt.subplots(3, 1, figsize=figsize)

    # Main plot
    f0 = ax[0].imshow(L, origin="lower", cmap='viridis', extent=subplot_extent,
        vmin=0, vmax=nanpercentile(L, 99))
    f1 = ax[1].imshow(P, origin="lower", cmap='viridis', extent=subplot_extent,
        vmin=0, vmax=nanpercentile(P, 99))

    # Detections per frame
    ax[2].plot(np.arange(n_blocks)*frame_block_size, detections_per_frame,
        color='k')
    ax[2].set_ylabel("Detections\nper frame", fontsize=fontsize)
    ax[2].set_xlabel("Frame", fontsize=fontsize)
    ax[2].set_ylim((0, ax[2].get_ylim()[1]))
    ax[2].yaxis.set_tick_params(labelsize=tick_fontsize)
    ax[2].xaxis.set_tick_params(labelsize=tick_fontsize)

    # Color scale
    cbar = plt.colorbar(f0, ax=ax[0], shrink=0.75)
    if normalize:
        cbar.set_label("Naive\nprobability", fontsize=fontsize)
    else:
        cbar.set_label("Unnormalized\nnaive\nprobability", fontsize=fontsize)
    cbar.ax.tick_params(labelsize=fontsize)

    cbar = plt.colorbar(f1, ax=ax[1], shrink=0.75)
    if normalize:
        cbar.set_label("Posterior\nprobability", fontsize=fontsize)
    else:
        cbar.set_label("Unnormalized\nposterior\nprobability", fontsize=fontsize)
    cbar.ax.tick_params(labelsize=tick_fontsize)

    # Ticks and axis labels
    xticks = np.arange(n_blocks)
    while xticks.shape[0] > 10:
        xticks = xticks[0::2]
    xticklabels = [f"{x*frame_block_size}" for x in xticks]
    for j in range(2):
        ax[j].set_xticks(xticks * subplot_extent[1] / n_blocks)
        ax[j].set_xticklabels(xticklabels, fontsize=tick_fontsize)
        ax[j].set_xlabel("Frame", fontsize=fontsize)       
        add_log_scale_imshow(ax[j], diff_coefs, fontsize=tick_fontsize, side='y')
        ax[j].set_ylabel("Diffusion\ncoefficient\n($\mu$m$^{2}$ s$^{-1}$)",
            fontsize=fontsize)

    # Main figure title
    if suptitle is not None:
        fig.suptitle(suptitle, fontsize=fontsize)

    # Save
    plt.tight_layout()
    plt.savefig(out_png, dpi=600)
    plt.close()

def spatial_assignment_probability_plot(
    out_png: str,
    tracks: pd.DataFrame,
    assignment_probabilities: np.ndarray,
    assignment_likelihoods: np.ndarray,
    diff_coefs: np.ndarray,
    pixel_size_um: float,
    fontsize: int=10,
    spatial_bin_size: int=2,
    smooth_kernel: float=0.0,
    suptitle: str=None,
):
    """ Experimental plot. Show the prior and posterior mean diffusion
    coefficients as a function of space. 

    Produces a 3-axis plot. The first axis shows detection density, the 
    second shows the mean prior diffusion coefficient, and the third shows
    the mean posterior diffusion coefficient.

    args
    ----
        out_png                 :   output filename
        tracks                  :   pandas.DataFrame, detections
        assignment_probabilities:   2D numpy.ndarray of shape (n_diff_coefs, n_tracks)
        assignment_likelihoods  :   2D numpy.ndarray of shape (n_diff_coefs, n_tracks)
        diff_coefs              :   1D numpy.ndarray of shape (n_diff_coefs,)
        pixel_size_um           :   float, camera pixel size in microns
        fontsize                :   int
        spatial_bin_size        :   int, number of camera pixels per spatial bin
        smooth_kernel           :   float, smoothing kernel in microns
        suptitle                :   str, main figure title if desired
    """
    # Estimate the size of the FOV
    if len(tracks) > 0:
        fov_size = (
            int(np.ceil(tracks[PY].max())) + 1,
            int(np.ceil(tracks[PX].max())) + 1
        )
    else:
        fov_size = (1, 1)

    # Number of distinct diffusion coefficients
    n_dc = diff_coefs.shape[0]

    # Number of trajectories
    n_tracks = assignment_probabilities.shape[1]
    track_indices = np.arange(n_tracks).astype(np.int64)
    tracks = tracks[tracks[TRACK].isin(track_indices)].reset_index(drop=True)

    # Spatial bins
    ybins = np.arange(0, fov_size[0]+spatial_bin_size, spatial_bin_size)
    xbins = np.arange(0, fov_size[1]+spatial_bin_size, spatial_bin_size)

    # Scale smoothing kernel into bin units
    smooth_kernel = smooth_kernel / (pixel_size_um * spatial_bin_size)

    # Bin area in microns
    bin_area = (spatial_bin_size * pixel_size_um)**2

    # Number of detections in each bin
    H = np.histogram2d(tracks[PY], tracks[PX], bins=(xbins, ybins))[0].astype(np.float64)
    H = gaussian_filter(H, smooth_kernel)
    occupied = H > 0

    # Normalize over states for each trajectory
    assignment_probabilities = normalize_2d(assignment_probabilities, 0)
    assignment_likelihoods = normalize_2d(assignment_likelihoods, 0)

    # Mean prior diffusion coefficient in each spatial bin
    mean_diff_coef = pd.Series((assignment_likelihoods.T * diff_coefs).sum(axis=1),
        index=track_indices, name="mean_prior_diff_coef")
    prior_mean = np.histogram2d(tracks[PY], tracks[PX], bins=(xbins, ybins),
        weights=tracks[TRACK].map(mean_diff_coef))[0].astype(np.float64)
    prior_mean = gaussian_filter(prior_mean, smooth_kernel)
    prior_mean[occupied] = prior_mean[occupied] / H[occupied]

    # Mean posterior diffusion coefficient in each spatial bin
    mean_diff_coef = pd.Series((assignment_probabilities.T * diff_coefs).sum(axis=1),
        index=track_indices, name="mean_posterior_diff_coef")
    post_mean = np.histogram2d(tracks[PY], tracks[PX], bins=(xbins, ybins),
        weights=tracks[TRACK].map(mean_diff_coef))[0].astype(np.float64)
    post_mean = gaussian_filter(post_mean, smooth_kernel)
    post_mean[occupied] = post_mean[occupied] / H[occupied]

    # Get localization density
    loc_density = H / bin_area 

    vmin = 0
    vmax = np.mean([np.nanpercentile(post_mean, 99), np.nanpercentile(prior_mean, 99)])

    # Main plot
    fig, ax = plt.subplots(1, 3, figsize=(3*3, 1*3), sharex=True, sharey=True)
    f0 = ax[0].imshow(loc_density, cmap='gray', origin='lower', vmin=vmin, vmax=vmax)
    cbar = plt.colorbar(f0, ax=ax[0], fraction=0.046, pad=0.04)
    cbar.set_label("Detections / $\mu$m$^{2}$", fontsize=fontsize)

    f1 = ax[1].imshow(prior_mean, cmap='viridis', origin='lower', vmin=vmin, vmax=vmax)
    cbar = plt.colorbar(f1, ax=ax[1], fraction=0.046, pad=0.04)
    cbar.set_label("Naive mean diff. coef. ($\mu$m$^{2}$ s$^{-1}$)", fontsize=fontsize)

    f2 = ax[2].imshow(post_mean, cmap='viridis', origin='lower', vmin=vmin, vmax=vmax)
    cbar = plt.colorbar(f2, ax=ax[2], fraction=0.046, pad=0.04)
    cbar.set_label("Posterior mean diff. coef. ($\mu$m$^{2}$ s$^{-1}$)", fontsize=fontsize)

    # Formatting
    for j in range(3):
        kill_ticks(ax[j], spines=False)

    # Main figure title
    if suptitle is not None:
        fig.suptitle(suptitle, fontsize=fontsize)

    plt.tight_layout()
    plt.savefig(out_png, dpi=600)
    plt.close()

#########################
## DATASET-LEVEL PLOTS ##
#########################

def heatmap_by_condition(
    out_png: str,
    diff_coefs: np.ndarray,
    dists: dict,
    order: List[str]=[],
    cmap: str="viridis",
    vmin: float=0,
    vmax: float=None,
    subplot_extent: Tuple[int]=(0, 5, 0, 1.5),
    ylabel: str=None,
    cbar_label: str=None,
    suptitle: str=None,
    fontsize: int=8,
    normalize: bool=False,
    order_by_size: bool=False,
):
    """ Plot groups of one-dimensional signals corresponding to different
    experimental conditions as a heat map. The signals are assumed to be
    defined on a logarithmically-spaced support of diffusion coefficients.

    args
    ----
        out_png         :   output file
        diff_coefs      :   1D numpy.ndarray of shape (n_diff_coefs,)
        dists           :   dict keyed by condition name to 2D numpy.ndarray.
                            Each array has shape (n_files, n_diff_coefs),
                            which gives each of the 1D signals corresponding
                            to that condition.
        order           :   list of str, order in which to plot the
                            different experimental conditions
        cmap            :   color map name
        vmin            :   color LUT min
        vmax            :   color LUT max
        subplot_extent  :   dimensions of individual subplots
        ylabel          :   y-axis label   
        cbar_label      :   color bar label
        suptitle        :   main figure title
        normalize       :   normalize each signal for unit intensity
        order_by_size   :   order the signals corresponding to each condition
                            by summed intensity (before normalization)
    """
    # Check compatibility of arguments
    if not all([
        isinstance(dists, dict),
        isinstance(diff_coefs, np.ndarray),
        len(diff_coefs.shape) == 1,
        all(map(lambda a: isinstance(a, np.ndarray), dists.values())),
        all(map(lambda a: a.shape[1] == diff_coefs.shape[0], dists.values()))
    ]):
        raise ValueError("incompatible arguments")

    if len(order) == 0:
        order = list(dists.keys())
        order.sort()
    elif not all(map(lambda c: c in dists.keys(), order)):
        raise ValueError(f"all elements of *order* must be keys in *dists*")

    # Number of conditions
    n_conditions = len(dists)

    # Number of files
    n_files = sum(map(lambda v: v.shape[0], dists.values()))

    # Normalize and order by increasing density, if relevant
    if order_by_size:
        for k in dists.keys():
            i = np.argsort(dists[k].sum(axis=1))
            dists[k] = dists[k][i[::-1],:]
    if normalize:
        for k in dists.keys():
            dists[k] = normalize_2d(dists[k], 1)

    # Color map scaling
    if vmin is None:
        vmin = vmin(map(np.nanmin, dists.values()))
    if vmax is None:
        vmax = nanpercentile(np.concatenate(list(dists.values()), axis=0), 99)

    # Plot layout. We favor stacking conditions vertically rather than
    # horizontally, but otherwise attempt to keep the subplots as square as
    # possible
    ny = int(np.ceil(np.sqrt(n_conditions)))
    nx = int(np.ceil(n_conditions / ny))

    fig, axes = plt.subplots(ny, nx, sharex=True, 
        figsize=(nx*subplot_extent[1], ny*subplot_extent[3]))
    axes = np.asarray(axes).reshape((ny, nx))
    axes = axes.ravel()

    def make_subplot(i: int):
        ax = axes[i]
        name = order[i]
        f = ax.imshow(dists[name], cmap=cmap, vmin=vmin, vmax=vmax, 
            extent=subplot_extent)
        cbar = plt.colorbar(f, ax=ax, fraction=0.046, pad=0.04)
        cbar.set_ticks([0])
        cbar.ax.tick_params(labelsize=fontsize)
        cbar.set_label(cbar_label, fontsize=fontsize)
        add_log_scale_imshow(ax, diff_coefs, fontsize=fontsize, side='x')
        if name != "no_condition":
            ax.set_title(name, fontsize=fontsize)
        ax.set_yticks([])
        if i // nx == ny - 1:
            ax.set_xlabel("Diff. coef. ($\mu$m$^{2}$ s$^{-1}$)", fontsize=fontsize)
        if ylabel is not None:
            ax.set_ylabel(ylabel, fontsize=fontsize)

    for i in range(n_conditions):
        make_subplot(i)

    # Empty subplots
    for i in range(n_conditions, ny*nx):
        f = axes[i].imshow(np.full(list(dists.values())[1].shape, np.nan), cmap=cmap,
            vmin=vmin, vmax=vmax, extent=subplot_extent)
        cbar = plt.colorbar(f, ax=axes[i], fraction=0.046, pad=0.04)
        cbar.set_ticks([0])
        cbar.ax.tick_params(labelsize=fontsize)
        cbar.set_label(cbar_label, fontsize=fontsize)
        add_log_scale_imshow(axes[i], diff_coefs, fontsize=fontsize, side='x')
        axes[i].set_yticks([])
        if i // nx == ny - 1:
            axes[i].set_xlabel("Diff. coef. ($\mu$m$^{2}$ s$^{-1}$)", fontsize=fontsize)
        for s in ['top', 'left', 'right']:
            axes[i].spines[s].set_visible(False)

    plt.tight_layout()
    plt.savefig(out_png, dpi=600)
    plt.close()

def lineplot_by_condition(
    out_png: str,
    diff_coefs: np.ndarray,
    dists: dict,
    order: List[str]=[],
    normalize: bool=True,
    fontsize: int=10,
    ylabel: str=None,
    alpha: float=0.5,
):
    # Check compatibility of arguments
    if not all([
        isinstance(dists, dict),
        isinstance(diff_coefs, np.ndarray),
        len(diff_coefs.shape) == 1,
        all(map(lambda a: isinstance(a, np.ndarray), dists.values())),
        all(map(lambda a: a.shape[1] == diff_coefs.shape[0], dists.values()))
    ]):
        raise ValueError("incompatible arguments")

    if len(order) == 0:
        order = list(dists.keys())
        order.sort()
    elif not all(map(lambda c: c in dists.keys(), order)):
        raise ValueError(f"all elements of *order* must be keys in *dists*")

    # Number of conditions
    n_conditions = len(dists)

    # Number of files
    n_files = sum(map(lambda v: v.shape[0], dists.values()))

    # Normalize
    if normalize:
        for k in dists.keys():
            dists[k] = normalize_2d(dists[k], axis=1)

    # Plot layout. We favor stacking conditions vertically rather than
    # horizontally, but otherwise attempt to keep the subplots as square as
    # possible
    ny = int(np.ceil(np.sqrt(n_conditions)))
    nx = int(np.ceil(n_conditions / ny))

    fig, axes = plt.subplots(ny, nx, sharex=True, sharey=True, figsize=(nx*4, ny*1.5))
    colors = sns.color_palette("dark", n_conditions)
    axes = np.asarray(axes).reshape((ny, nx))
    axes = axes.ravel()

    def make_subplot(i: int):
        ax = axes[i]
        name = order[i]
        for j in range(dists[name].shape[0]):
            ax.plot(diff_coefs, dists[name][j,:], color=colors[i], label=None, alpha=alpha)
        ax.plot([], [], color=colors[i], alpha=alpha, label="File")
        ax.plot(diff_coefs, np.nanmean(dists[name], axis=0), color='k', linestyle='--', 
            label="Mean")

        ax.legend(frameon=False, loc='upper right', prop={'size': 6})
        ax.set_title(name, fontsize=fontsize)
        for s in ['top', 'right']:
            ax.spines[s].set_visible(False)

        # y-ticks
        ax.set_yticks([])
        if (ylabel is not None) and (i % nx == 0):
            ax.set_ylabel(ylabel, fontsize=fontsize)
        ax.set_ylim((0, ax.get_ylim()[1]))

        # x-ticks
        ax.set_xscale("log")
        if i // nx == ny - 1:
            ax.set_xlabel("Diff. coef. ($\mu$m$^{2}$ s$^{-1}$)", fontsize=fontsize)

    for i in range(n_conditions):
        make_subplot(i)

    for i in range(n_conditions, ny*nx):
        for s in ['top', 'left', 'right']:
            axes[i].spines[s].set_visible(False)
        if i // nx == ny - 1:
            axes[i].set_xlabel("Diff. coef. ($\mu$m$^{2}$ s$^{-1}$)", fontsize=fontsize)

    # Save
    plt.tight_layout()
    plt.savefig(out_png, dpi=600)
    plt.close()






import os, warnings, numpy as np, pandas as pd 
from tqdm import tqdm
from scipy.special import digamma
from typing import Tuple

from .constants import (
    DEFAULT_MAX_ITER, DEFAULT_CONC_PARAM, DEFAULT_SPLITSIZE,
    DEFAULT_START_FRAME, RBME, RBME_MARGINAL, GAMMA, FBME, FRAME
)
from .trajectory_group import TrajectoryGroup
from .lik import Likelihood, make_likelihood, make_likelihood_from_params
from .parameters import StateArrayParameters
from .plot import (
    rbm_posterior_plot,
    rbme_posterior_plot,
    marginal_assignment_probability_plot,
    temporal_assignment_probability_plot,
    spatial_assignment_probability_plot,
)
from .utils import cartesian_product

class StateArray:
    """ Represents a state array for a particular set of trajectories and a 
    particular likelihood function.

    A state array is a mixture model. Trajectories are assumed to inhabit one
    of a "grid" of states with distinct parameters. The StateArray object
    provides methods to infer the posterior distribution over trajectory-state
    assignments and state occupations, given an observed set of trajectories.

    init
    ----
        trajectories    :   observed set of trajectories in an SPT experiment
        likelihood      :   the likelihood function with respect to which the
                            state array is defined
        params          :   settings for inference

    attributes
    ----------
        n_tracks                :   number of observed trajectories

        n_jumps                 :   number of observed jumps (spot-spot
                                    connections)

        n_detections            :   number of observed spots

        shape                   :   shape of the parameter grid on which the
                                    state occupations are defined

        jumps_per_track         :   number of observed jumps per trajectory

        naive_assignment_probabilities    : naive probabilities for each 
                                    trajectory-state assignment

        posterior_assignment_probabilities: posterior probability for each 
                                    trajectory-state assignment

        prior_occs              :   mean occupation of each state under the
                                    prior distribution over state occupations

        naive_occs              :   mean occupation of each state under the
                                    naive distribution over state occupations

        posterior_occs          :   mean occupation of each state under the 
                                    posterior distribution over state occupations 

        prior_dirichlet_param   :   parameter to the prior Dirichlet distribution
                                    over state occupations

        posterior_dirichlet_param:  parameter to the posterior Dirichlet distribution
                                    over state occupations
    """
    def __init__(self, trajectories: TrajectoryGroup,
        likelihood: Likelihood, params: StateArrayParameters):
        self.trajectories = trajectories
        self.likelihood = likelihood
        self.params = params

    def __repr__(self):
        """ String representation of this StateArray object """    
        return "StateArray:\n  {}".format("\n  ".join([
            f"{a : <18}: {getattr(self, a)}" for a in \
                ["likelihood_type", "n_tracks", "n_jumps", "parameter_names", "shape"]
        ]))

    def __enter__(self):
        return self 

    def __exit__(self, etype, evalue, traceback):
        return etype is None

    @classmethod
    def from_detections(cls, detections: pd.DataFrame, likelihood_type: str, **kwargs):
        """ Convenience method; instantiate a StateArray object by passing 
        a raw set of detections as a pandas.DataFrame.

        args
        ----
            detections          :   pandas.DataFrame indexed by detection with
                                    columns TRACK, FRAME, PY, and PX at minimum
            likelihood_type     :   type of likelihood function to use
            **kwargs            :   any parameters supported by StateArrayParameters

        returns
        -------
            new instance of StateArray
        """
        params = StateArrayParameters(**kwargs)
        trajectories = TrajectoryGroup.from_params(detections, params)
        if params.sample_size < trajectories.n_tracks:
            trajectories = trajectories.subsample(params.sample_size)
        likelihood = make_likelihood_from_params(likelihood_type, params, **kwargs)
        return cls(trajectories, likelihood, params)

    ################
    ## PROPERTIES ##
    ################

    @property 
    def n_tracks(self) -> int:
        return self.trajectories.n_tracks 

    @property 
    def n_jumps(self) -> int:
        return self.trajectories.n_jumps 

    @property 
    def n_detections(self) -> int:
        return self.trajectories.n_detections

    @property 
    def shape(self) -> Tuple[int]:
        return self.likelihood.shape

    @property 
    def likelihood_type(self) -> str:
        return self.likelihood.name

    @property 
    def parameter_names(self) -> str:
        return self.likelihood.parameter_names

    @property 
    def parameter_values(self) -> Tuple[np.ndarray]:
        return self.likelihood.parameter_values

    @property
    def n_states(self) -> int:
        """ Total number of states in the state array """
        if not hasattr(self, "_n_states"):
            self._n_states = np.prod(self.shape)
        return self._n_states

    @property
    def jumps_per_track(self) -> np.ndarray:
        """ Number of jumps per trajectory.

        returns
        -------
            1D numpy.ndarray of shape (n_tracks,)
        """
        if not hasattr(self, "_jumps_per_track"):
            self._calc_assignment_likelihoods()
        return self._jumps_per_track

    @property
    def naive_assignment_probabilities(self):
        """ Naive probability for each trajectory-state assignment.

        This is really just the likelihood function normalized over all
        states for each trajectory. It is equivalent to the posterior
        probability the trajectory-state assignments if

            a. we assume a uniform prior over trajectory-state assignments

            b. we condition on uniform state occupations; that is, the
                occupations of all states in the parameter grid are 
                identical

        returns
        -------
            numpy.ndarray of shape (*self.shape, n_tracks)
        """       
        if not hasattr(self, "_naive_assignment_probabilities"):
            self._calc_assignment_likelihoods()
        return self._naive_assignment_probabilities

    @property 
    def posterior_assignment_probabilities(self) -> np.ndarray:
        """ Posterior probability for each trajectory-state assignment.

        returns
        -------
            numpy.ndarray of shape (*self.shape, n_tracks)
        """
        if not hasattr(self, "_posterior_assignment_probabilities"):
            self._infer_posterior()
        return self._posterior_assignment_probabilities

    @property
    def prior_dirichlet_param(self) -> np.ndarray:
        """ Parameter for the prior Dirichlet distribution over state
        occupations. 

        returns
        -------
            numpy.ndarray of shape *self.shape*
        """
        if not hasattr(self, "_prior_dirichlet_param"):
            self._prior_dirichlet_param = np.full(self.shape,
                self.params.conc_param, dtype=np.float64)
        return self._prior_dirichlet_param

    @property
    def posterior_dirichlet_param(self) -> np.ndarray:
        """ Parameter for the posterior Dirichlet distribution over state
        occupations.

        returns
        -------
            numpy.ndarray of shape *self.shape*.
        """
        if not hasattr(self, "_posterior_dirichlet_param"):
            self._infer_posterior()
        return self._posterior_dirichlet_param

    @property
    def prior_occs(self) -> np.ndarray:
        """ Mean of the prior Dirichlet distribution over state occupations.

        returns
        -------
            numpy.ndarray of shape *self.shape*
        """
        if not hasattr(self, "_prior_occs"):
            if self.params.conc_param <= 0.0:
                self._prior_occs = np.ones(self.shape, dtype=np.float64)
                if self._prior_occs.size > 0:
                    self._prior_occs /= self._prior_occs.sum()
            else:
                S = self.prior_dirichlet_param.sum()
                if S > 0 and self.prior_dirichlet_param.size > 0:
                    self._prior_occs = self.prior_dirichlet_param / S 
                else:
                    self._prior_occs = np.ones(self.shape, dtype=np.float64)
        return self._prior_occs

    @property
    def naive_occs(self) -> np.ndarray:
        """ Naive estimate for the occupations of each state. This is obtained
        by marginalizing the trajectory-state assignment likelihoods over all
        trajectories.

        Normalized, so that naive_occs.sum() == 1.0.

        returns
        -------
            numpy.ndarray of shape *self.shape*
        """
        if not hasattr(self, "_naive_occs"):
            # Scale naive assignment likelihoods by the number of jumps
            # in each trajectory, and marginalize over the assignments
            self._naive_occs = (self.naive_assignment_probabilities * \
                self.jumps_per_track).sum(axis=-1)

            # Apply an appropriate defocalization correction
            self._naive_occs = self.likelihood.correct_for_defocalization(self._naive_occs)

            # Normalize over all states
            if self._naive_occs.sum() > 0:
                self._naive_occs /= self._naive_occs.sum()
        return self._naive_occs

    @property 
    def posterior_occs(self) -> np.ndarray:
        """ Mean of the posterior Dirichlet distribution over state occupations.

        returns
        -------
            numpy.ndarray of shape *self.shape*.
        """
        if not hasattr(self, "_posterior_occs"):
            self._infer_posterior()
        return self._posterior_occs

    @property
    def occupations_dataframe(self) -> pd.DataFrame:
        """ pandas.DataFrame representation of the posterior distribution,
        suitable for saving to a CSV.

        In this representation, each row of the DataFrame corresponds to a 
        distinct state in the state array. We give the marginal likelihood 
        and mean posterior occupation for each state. """
        if not hasattr(self, "_occupations_dataframe"):
            parameters = list(self.likelihood.parameter_names)
            cols = parameters + ["naive_occupation", "mean_posterior_occupation"]
            df = pd.DataFrame(index=np.arange(self.n_states), columns=cols, dtype=np.float64)
            df["naive_occupation"] = self.naive_occs.ravel()
            df["mean_posterior_occupation"] = self.posterior_occs.ravel()
            if self.n_states > 0:
                df[parameters] = cartesian_product(*self.likelihood.parameter_values)
            self._occupations_dataframe = df
        return self._occupations_dataframe

    @property 
    def diff_coefs(self) -> np.ndarray:
        """ Convenience attribute; the set of diffusion coefficients
        supported by the underlying Likelihood function.

        returns
        -------
            1D numpy.ndarray, the diffusion coefficients in µm2/sec
        """
        if not hasattr(self, "_diff_coefs"):
            if "diff_coef" in self.likelihood.parameter_names:
                self._diff_coefs = self.likelihood.diff_coefs
            else:
                self._diff_coefs = np.array([], dtype=np.float64)
        return self._diff_coefs

    #############
    ## METHODS ##
    #############

    def infer_posterior(self) -> Tuple[np.ndarray]:
        """ Attempt to infer the posterior distribution over
        trajectory-state assignments and state occupations for
        this StateArray.

        returns
        -------
            (
                numpy.ndarray of shape (*self.shape, n_tracks), 
                    posterior probabilities for each trajectory-state
                    assignment;

                numpy.ndarray of shape *self.shape*, parameter
                    to the posterior Dirichlet distribution over state
                    occupations;

                numpy.ndarray of shape *self.shape*, mean of the
                    posterior Dirichlet distribution over state occupations
            )
        """
        L = self.naive_assignment_probabilities
        R = L.copy()
        par_indices = tuple(range(len(self.shape)))
        if self.params.progress_bar:
            print("inferring posterior distribution...")
            iterations = tqdm(range(self.params.max_iter))
        else:
            iterations = range(self.params.max_iter)
        for i in iterations:

            # Update posterior state occupations (*n* is the parameter to 
            # a Dirichlet distribution over state occupations)
            n = (R * self.jumps_per_track).sum(axis=-1)
            m = n + self.prior_dirichlet_param

            # Exponent of the expected log occupation under the current
            # posterior estimate
            exp_log_tau = np.exp(digamma(m))

            # Calculate posterior probabilities for each trajectory-state
            # assignment, normalizing over all states for each trajectory
            R = (L.T * exp_log_tau.T).T
            R = R / R.sum(axis=par_indices)

        # Special case: 0 iterations
        if self.params.max_iter == 0:
            n = (R * self.jumps_per_track).sum(axis=-1)

        # Adjust posterior distribution to account for defocalization
        n = self.likelihood.correct_for_defocalization(n, normalize=False)

        # Calculate mean state occupations under the posterior model
        post_mean = n / n.sum() if n.sum() > 0 else self.prior_occs.copy()

        return R, n, post_mean

    def marginalize_on_diff_coef(self, *args, **kwargs):
        """ Inherited from the underlying Likelihood object. """
        return self.likelihood.marginalize_on_diff_coef(*args, **kwargs)

    ##############
    ## PLOTTING ##
    ##############

    def plot_occupations(self, out_png: str, **kwargs):
        """ Make a plot of the posterior distribution suitable for the 
        underlying likelihood function of this StateArray.

        May not be implemented for all likelihood functions.

        args
        ----
            out_png     :   output path for the plot
            kwargs      :   additional keyword arguments to the plotting
                            function
        """
        if self.likelihood.name == RBME:
            rbme_posterior_plot(out_png, self.naive_occs, 
                self.posterior_occs, self.likelihood.diff_coefs,
                self.likelihood.loc_errors, **kwargs)
        elif self.likelihood.name in [RBME_MARGINAL, GAMMA]:
            rbm_posterior_plot(out_png, self.naive_occs, 
                self.posterior_occs, self.likelihood.diff_coefs,
                **kwargs)
        else:
            warnings.warn(f"no posterior plot is implemented for " \
                "likelihood {self.likelihood.name}; not making plot")

    def plot_assignment_probabilities(self, out_png: str, **kwargs):
        """ Make a plot of the posterior probabilities for each
        trajectory-state assignment. 

        The likelihood function must support the diffusion coefficient
        parameter.

        args
        ----
            out_png     :   output path for plot
            kwargs      :   additional keyword arguments to the plotting
                            function
        """
        if "diff_coef" not in self.likelihood.parameter_names:
            warnings.warn(f"no posterior assignment plot is implemented " \
                "for likelihood {self.likelihood.name}; not making plot")
        else:
            scaled_assignment_probabilities = self.naive_assignment_probabilities * \
                self.jumps_per_track
            marginal_assignment_probability_plot(
                out_png,
                self.likelihood.marginalize_on_diff_coef(self.posterior_assignment_probabilities),
                self.likelihood.marginalize_on_diff_coef(scaled_assignment_probabilities),
                self.likelihood.diff_coefs,
                sort_by_mean_diff_coef=True,
                **kwargs
            )

    def plot_temporal_assignment_probabilities(self, out_png: str,
        frame_block_size: int=None, **kwargs):
        """ Make a plot of the posterior probabilities for each 
        trajectory-state assignment as a function of the frame index.

        Useful for understanding whether density (which usually begins
        high and declines throughout an SPT experiment) has a noticeable
        effect on the trajectory-state probabilities.

        args
        ----
            out_png             :   output path for plot
            frame_block_size    :   temporal bin size in frames
            kwargs              :   additional keyword arguments to
                                    plot function
        """
        if "diff_coef" not in self.likelihood.parameter_names:
            warnings.warn(f"no posterior assignment plot is implemented " \
                "for likelihood {self.likelihood.name}; not making plot")
        else:
            P = self.likelihood.marginalize_on_diff_coef(self.posterior_assignment_probabilities)
            L = self.likelihood.marginalize_on_diff_coef(self.naive_assignment_probabilities)
            P = P * self.jumps_per_track
            L = L * self.jumps_per_track

            # Choose a frame block size appropriate for this movie
            if frame_block_size is None:
                n_frames = self.trajectories.detections[FRAME].max() + 1
                base = max(0, np.log10(n_frames) - 2)
                exp_base = 10**int(np.floor(base))
                frame_block_size = exp_base if (base%1.0<0.5) else 3*exp_base

            temporal_assignment_probability_plot(
                out_png,
                self.trajectories.detections,
                P,
                L,
                self.likelihood.diff_coefs,
                frame_block_size=frame_block_size,
                **kwargs
            )

    def plot_spatial_assignment_probabilities(self, out_png: str, **kwargs):
        """ Aggregate the prior and posterior trajectory-state assignment probabilities
        into spatial bins and show alongside the localization density.

        args
        ----
            out_png     :   output path for plot
            kwargs      :   additional keyword arguments to plot function
        """
        if "diff_coef" not in self.likelihood.parameter_names:
            warnings.warn(f"no posterior assignment plot is implemented " \
                "for likelihood {self.likelihood.name}; not making plot")
        else:
            P = self.likelihood.marginalize_on_diff_coef(self.posterior_assignment_probabilities)
            L = self.likelihood.marginalize_on_diff_coef(self.naive_assignment_probabilities)
            spatial_assignment_probability_plot(
                out_png,
                self.trajectories.detections,
                P,
                L,
                self.likelihood.diff_coefs,
                self.params.pixel_size_um,
                **kwargs
            )

    ######################
    ## OBJECT UTILITIES ##
    ######################

    def _infer_posterior(self):
        self._posterior_assignment_probabilities, \
            self._posterior_dirichlet_param, \
            self._posterior_occs = self.infer_posterior()

    def _calc_assignment_likelihoods(self) -> Tuple[np.ndarray]:
        """ Calculate the normalized likelihood and number of jumps for each 
        of the trajectories in this state array. """
        log_L, self._jumps_per_track = self.likelihood(self.trajectories)
        self._naive_assignment_probabilities = self.likelihood.exp(log_L)



#!/usr/bin/env python
__version__ = "0.3.0"
from .constants import RBME, RBME_MARGINAL, GAMMA, FBME, LIKELIHOOD_TYPES
from .dataset import StateArrayDataset
from .io import load_detections, concat_detections, sample_detections
from .lik import make_likelihood, LIKELIHOODS
from .parameters import StateArrayParameters
from .sa import StateArray
from .trajectory_group import TrajectoryGroup
from .utils import normalize_2d



#!/usr/bin/env python
"""
defoc.py -- evaluate defocalization probabilities for various diffusion models
"""
import os 
import numpy as np 
import pandas as pd 

# Caching
from functools import lru_cache

# Cubic spline interpolation, for FBM defocalization function
from scipy import interpolate 

# Names of likelihood function
from .constants import RBME, GAMMA, RBME_MARGINAL, FBME

# saspt module directory
PACKAGE_DIR = os.path.split(os.path.abspath(__file__))[0]

# Directory with spline coefficients for FBM defocalization
SPLINE_DIR = os.path.join(PACKAGE_DIR, "splines")


###############
## UTILITIES ##
###############


@lru_cache(maxsize=1)
def load_fbm_defoc_spline(dz=0.7):
    """
    Given a focal depth, get a spline interpolator that enables calculation
    of the fraction of FBMs that defocalize at various frame intervals.
    args
    ----
        dz      :   float, the focal depth in um
    returns
    -------
        5-tuple, the *tck* argument expected by scipy.interpolate's spline
            evaluators -- specifically scipy.interpolate.bisplev
    """
    # Available frame intervals
    avail_dz = np.array([0.1, 0.2, 0.3, 0.4, 0.5, 0.6, 0.7, 0.8, 0.9, 1.0,
        1.1, 1.2, 1.3, 1.4, 1.5, 1.6])

    # Get the closest available focal depth
    m = np.argmin(np.abs(avail_dz - dz))
    sel_dz = avail_dz[m]

    # Path to this file
    path = os.path.join(SPLINE_DIR, "fbm_defoc_splines_dz-%.1f.csv" % sel_dz)

    # Load the spline coefficients
    tcks = load_spline_coefs_multiple_frame_interval(path)
    return tcks


def load_spline_coefs_multiple_frame_interval(path):
    """
    Load multiple sets of bivariate spline coefficients from a file.
    These are in the format required by scipy.interpolate for 
    evaluation of bivariate splines.
    The individual sets of spline coefficients are ;-delimited, while
    the different parts of the coefficient 5-tuple are newline-delimited and
    the individual numbers are ,-delimited.
    args
    ----
        path        :   str, path to a file of the type written by
                        save_spline_coefs_multiple()
    returns
    -------
        list of 5-tuple, the bivariate spline coefficients for each
            frame interval
    """
    with open(path, "r") as f:
        S = f.read().split(";")
    S = [j.split("\n") for j in S]
    result = []
    for lines in S:
        x = np.asarray([float(j) for j in lines[0].split(",")])
        y = np.asarray([float(j) for j in lines[1].split(",")])
        coefs = np.array([float(j) for j in lines[2].split(",")])
        kx = int(lines[3])
        ky = int(lines[4])
        result.append((x, y, coefs, kx, ky))
    return result 


def eval_spline(x, y, tck):
    """
    Evaluate a bivariate spline on the Cartesian product of a set of X points
    and a set of Y points.
    args
    ----
        x       :   1D ndarray, the array of unique x points
        y       :   1D ndarray, the array of unique y points
        tck     :   5-tuple, bivariate spline coefficients of the type
                    read by *load_spline_coefs*
    returns
    -------
        2D ndarray of shape (y.shape[0], x.shape[0]), the evaluated
            bivariate spline at each combination of the input points
    """
    return interpolate.bisplev(x, y, tck).T 


def f_remain_rbm(D, n_frames, frame_interval, dz):
    """
    Calculate the fraction of regular Brownian particles that 
    remain in a microscope's depth of field after some number 
    of frames.
    args
    ----
        D               :   float, diffusion coefficient
                            in um^2 s^-1
        n_frames        :   int, the number of frames
        frame_interval  :   float, seconds
        dz              :   float, depth of field in um
    returns
    -------
        1D ndarray of shape (n_frames,), the probability
            to remain at each frame interval
    """
    if (dz is np.inf) or (dz is None) or (D <= 0.0):
        return np.ones(n_frames, dtype=np.float64)

    # Support for the calculations
    s = (int(dz//2.0)+1) * 2
    support = np.linspace(-s, s, int(((2*s)//0.001)+2))[:-1]
    hz = 0.5 * dz 
    inside = np.abs(support) <= hz 
    outside = ~inside 

    # Define the transfer function for this BM
    g = np.exp(-(support ** 2)/ (4 * D * frame_interval))
    g /= g.sum()
    g_rft = np.fft.rfft(g)   

    # Set up the initial probability density
    pmf = inside.astype(np.float64)
    pmf /= pmf.sum()

    # Propagate over subsequent frame intervals
    result = np.zeros(n_frames, dtype=np.float64)
    for t in range(n_frames):
        pmf = np.fft.fftshift(np.fft.irfft(
            np.fft.rfft(pmf) * g_rft, n=pmf.shape[0]))
        pmf[outside] = 0.0
        result[t] = pmf.sum()

    return result 


def f_remain_fbm(D, hurst, n_frames, frame_interval, dz, D_type=4):
    """
    Calculate the fraction of fractional Brownian particles that 
    remain in a microscope's depth of field after some number of 
    frames.
    args
    ----
        D               :   float, diffusion coefficient
                            in um^2 s^-1
        hurst           :   float, Hurst parameter
        n_frames        :   int, the number of frames
        frame_interval  :   float, seconds
        dz              :   float, depth of field in um
    returns
    -------
        1D ndarray of shape (n_frames,), the probability
            to remain at each frame interval
    """
    if (dz is np.inf) or (dz is None) or (D <= 0.0):
        return np.ones(n_frames, dtype=np.float64)

    if n_frames > 8:
        raise RuntimeError("no more than 8 frame intervals supported " \
            "for FBM defocalization")

    # Get the dispersion parameter
    if D_type == 1:
        c = np.log(D * np.power(frame_interval, 2*hurst))
    elif D_type == 2:
        c = 2 * hurst * np.log(D * frame_interval)
    elif D_type == 3:
        c = np.log(D * frame_interval / (2 * hurst))
    elif D_type == 4:
        c = np.log(D * frame_interval)

    # Load spline coefficients for this focal depth
    tcks = load_fbm_defoc_spline(dz=dz)

    # Evaluate the probability of defocalization at each frame interval
    return np.asarray([eval_spline(hurst, c, tck) for tck in tcks[:n_frames]])


##############################################
## LIKELIHOOD-SPECIFIC CORRECTION FUNCTIONS ##
##############################################


def defoc_corr_rbm(L, diff_coefs, mode="point", frame_interval=0.00748, dz=0.7,
    normalize=False):
    """
    Apply a defocalization correction to the regular Brownian motion
    likelihood function.
    args
    ----
        L               :   ndarray, the likelihood function as generated
                            by *eval_likelihood*
        diff_coefs      :   1D ndarray, the set of diffusion coefficients
                            corresponding to *L*. The second axis of *L*
                            is assumed to corresponding to the values of 
                            *diff_coefs*.
        mode            :   str, the mode the likelihood was calculated in,
                            either "point" or "binned"
        frame_interval  :   float, frame interval in seconds
        dz              :   float, focal depth in microns
    returns
    -------
        reference to *L* after correction
    """
    diff_coefs = np.asarray(diff_coefs)
    if mode == "point":
        K = diff_coefs.shape[0]

        # For each diffusion coefficient, evaluate the defocalization
        # probability at one frame interval
        frac_remain = np.ones(K, dtype=np.float64)
        if not np.isinf(dz):
            for i, D in enumerate(diff_coefs):
                frac_remain[i] = f_remain_rbm(D, 1, frame_interval, dz)[0]       

    elif mode == "binned":
        K = diff_coefs.shape[0] - 1
        d_mid = np.sqrt(diff_coefs[1:] * diff_coefs[:-1])

        # For each diffusion coefficient, evaluate the defocalization
        # probability at one frame interval
        frac_remain = np.ones(K, dtype=np.float64)
        if not np.isinf(dz):
            for i, D in enumerate(d_mid):
                frac_remain[i] = f_remain_rbm(D, 1, frame_interval, dz)[0]

    # Passed a single aggregate likelihood function
    if len(L.shape) == 1:
        assert L.shape[0] == K, "first axis of aggregate likelihood must " \
            "correspond to the diffusion coefficient"

        # Apply the correction and renormalize
        L /= frac_remain
        if normalize:
            norm = L.sum()
            if norm > 0:
                L /= norm 

    else:
        assert L.shape[0] == K, "first axis of likelihood matrix must " \
            "correspond to the diffusion coefficient"

        # Apply the correction and renormalize
        if len(L.shape) == 2:
            L = (L.T / frac_remain).T
            if normalize: L = L / L.sum(axis=0)

        elif len(L.shape) == 3:
            L = (L.T / frac_remain).T
            if normalize: L = L / L.sum(axis=(0,1))

    return L 


def defoc_corr_rbme(L, diff_coefs, loc_errors, frame_interval=0.00748, dz=0.7,
    normalize=False):
    """
    Apply a defocalization correction to the regular Brownian motion with
    error (RBME) likelihood function. Localization error is not assumed
    to figure into defocalization, so the *loc_errors* argument is ignored.
    args
    ----
        L               :   ndarray, the likelihood function as generated
                            by *eval_likelihood*
        diff_coefs      :   1D ndarray, the set of diffusion coefficients
                            corresponding to *L*. The second axis of *L*
                            is assumed to corresponding to the values of 
                            *diff_coefs*.
        loc_errors      :   1D ndarray, localization error parameters
        frame_interval  :   float, frame interval in seconds
        dz              :   float, focal depth in microns
    returns
    -------
        reference to *L* after correction
    """
    diff_coefs = np.asarray(diff_coefs)
    K = diff_coefs.shape[0]

    # For each diffusion coefficient, evaluate the defocalization
    # probability at one frame interval
    frac_remain = np.ones(K, dtype=np.float64)
    if not np.isinf(dz):
        for i, D in enumerate(diff_coefs):
            frac_remain[i] = f_remain_rbm(D, 1, frame_interval, dz)[0]

    # Passed a single aggregate likelihood function
    if len(L.shape) == 2:
        assert L.shape[0] == K, "first axis of aggregate likelihood must " \
            "correspond to the diffusion coefficient"

        # Apply the correction and renormalize
        L = (L.T / frac_remain).T
        if normalize:
            norm = L.sum()
            if norm > 0:
                L /= norm

    elif len(L.shape) == 3:
        assert L.shape[0] == K, "first axis of likelihood matrix must " \
            "correspond to the diffusion coefficient"
        L = (L.T / frac_remain).T
        if normalize: L = L / L.sum(axis=(0,1))

    return L 


def defoc_corr_fbm(L, diff_coefs, hurst_pars, loc_errors, frame_interval=0.00748, dz=0.7,
    normalize=False):
    """
    Apply a defocalization correction to the fractional Brownian motion 
    likelihood function.
    Both diffusion coefficient and Hurst parameter figure into the
    defocalization function for FBM.
    args
    ----
        L               :   ndarray, the likelihood function as generated
                            by *eval_likelihood*
        diff_coefs      :   1D ndarray, the set of diffusion coefficients
                            corresponding to *L*. The second axis of *L*
                            is assumed to correspond to the values of 
                            *diff_coefs*.
        hurst_pars      :   1D ndarray, the set of Hurst parameters 
                            corresponding to *L*. The third axis of *L*
                            is assumed to correspond to the values of 
                            *hurst_pars*.
        loc_errors      :   1D ndarray, the set of localization errors
                            corresponding to *L*. The fourth axis of *L*
                            is assumed to correspond to the values of
                            *loc_errors*
        frame_interval  :   float, frame interval in seconds
        dz              :   float, focal depth in microns
    returns
    -------
        reference to *L* after correction
    """
    diff_coefs = np.asarray(diff_coefs)
    hurst_pars = np.asarray(hurst_pars)
    loc_errors = np.asarray(loc_errors)
    nD = diff_coefs.shape[0]
    nH = hurst_pars.shape[0]
    nLE = loc_errors.shape[0]

    # Evaluate the defocalization probability at one frame interval
    # for all parameter combinations considered in this likelihood matrix
    frac_remain = np.ones((nD, nH, nLE), dtype=np.float64)
    if not np.isinf(dz):
        for i, D in enumerate(diff_coefs):
            for j, H in enumerate(hurst_pars):
                frac_remain[i,j,:] = f_remain_fbm(D, H, 1, frame_interval, dz)

    # Aggregate likelihood
    if len(L.shape) == 3:

        # Apply the correction
        L /= frac_remain

        # Renormalize
        if normalize:
            norm = L.sum()
            if norm > 0:
                L /= norm

    # Trajectory-wise
    elif len(L.shape) == 4:
        L = (L.T / frac_remain.T).T
        if normalize: L = L / L.sum(axis=(0,1,2))

    return L 


# Available likelihood corrections
LIKELIHOOD_CORR_FUNCS = {
    GAMMA        : defoc_corr_rbm,
    RBME_MARGINAL: defoc_corr_rbm,
    RBME         : defoc_corr_rbme,
    FBME         : defoc_corr_fbm,
}


def defoc_corr(L, support, likelihood, frame_interval, dz, normalize=False, **kwargs):
    """
    Apply a defocalization correction to a likelihood function.
    args
    ----
        L           :   ndarray, the likelihood function as generated
                        by *eval_likelihood*
        support     :   tuple of ndarray, the parameter values for the 
                        support of *L*
        likelihood  :   str, the type of likelihood function
        frame_interval: float, frame interval in seconds
        dz          :   float, focal depth in microns
        kwargs      :   additional kwargs to the defocalization function
                        specific to this likelihood
    returns
    -------
        reference to *L* after correction
    """
    # Check that the likelihood function is supported
    if not likelihood in LIKELIHOOD_CORR_FUNCS.keys():
        raise ValueError("likelihood {} not found in correction functions".format(likelihood))

    # Apply the correction
    return LIKELIHOOD_CORR_FUNCS[likelihood](L, *support, dz=dz,
        frame_interval=frame_interval, normalize=normalize, **kwargs)



#!/usr/bin/env python
import numpy as np

#############################
## DEFAULT PARAMETER GRIDS ##
#############################

DEFAULT_DIFF_COEFS = np.logspace(-2.0, 2.0, 100)
DEFAULT_LOC_ERRORS = np.arange(0.0, 0.072, 0.002)
DEFAULT_HURST_PARS = np.arange(0.05, 1.0, 0.05)

###############################################
## DEFAULT PREPROCESSING AND HYPERPARAMETERS ##
###############################################

# When trajectories are too long, split them into smaller trajectories.
# *splitsize* defines the maximum trajectory length (in # jumps) before
# splitting.
DEFAULT_SPLITSIZE = 10

# Default concentration parameter for the prior distribution over state
# occupations.
DEFAULT_CONC_PARAM = 1.0

# Default number of iterations to do when inferring posterior
DEFAULT_MAX_ITER = 200

# Default first frame to consider
DEFAULT_START_FRAME = 0

# Maximum number of trajectories to consider when running state arrays
DEFAULT_SAMPLE_SIZE = 10000

################################
## DETECTION-LEVEL ATTRIBUTES ##
################################

# Column in the detections DataFrame encoding frame index
FRAME = "frame"

# Column in the detections DataFrame encoding trajectory index
TRACK = "trajectory"

# Column in the detections DataFrame encoding trajectory length (in frames)
TRACK_LENGTH = "track_length"

# Column in the detections DataFrame encoding y-position in pixels
PY = "y"

# Column in the detections DataFrame encoding x-position in pixels
PX = "x"

###########################
## JUMP-LEVEL ATTRIBUTES ##
###########################

# Column in the jumps DataFrame encoding number of frames over which
# the jump happened
DFRAMES = "dframes"

# Column in the jumps DataFrame encoding the change in y-position in microns
DY = "dy"

# Column in the jumps DataFrame encoding the change in x-position in microns
DX = "dx"

# Column in the jumps DataFrame encoding the squared 2D radial jump length
# in squared microns
DR2 = "dr2"

# Column in the jumps DataFrame encoding the number of jumps per trajectory
JUMPS_PER_TRACK = "jumps_per_track"

####################################
## AVAILABLE LIKELIHOOD FUNCTIONS ##
####################################

# Names of likelihood functions
RBME = "rbme"
RBME_MARGINAL = "rbme_marginal"
GAMMA = "gamma"
FBME = "fbme"

# All available likelihood functions
LIKELIHOOD_TYPES = [RBME, RBME_MARGINAL, GAMMA, FBME]

###########
## OTHER ##
###########

# Bucket condition column for StateArrayDatasets without an experimental condition
DEFAULT_CONDITION_COL = "default_condition"
DEFAULT_CONDITION = "no_condition"



import warnings, numpy as np
from typing import Tuple
from abc import ABC, abstractmethod

from .constants import (
    DEFAULT_DIFF_COEFS, DEFAULT_LOC_ERRORS, DEFAULT_HURST_PARS,
    GAMMA, RBME_MARGINAL, RBME, FBME, TRACK, DR2
)
from .defoc import defoc_corr
from .trajectory_group import TrajectoryGroup
from .parameters import StateArrayParameters

class Likelihood(ABC):
    """ Base class for likelihood functions used in state arrays. Each subclass
    must define attributes that identify the likelihood function and methods 
    required for downstream use in state arrays.

    In more detail, state arrays require likelihood functions that evaluate on a
    discrete grid of parameter values. The Likelihood object passes information
    about its parameter grid to the StateArray object via the *shape*, 
    *parameter_names*, and *parameter_values* attributes. """
    @property
    @abstractmethod
    def name(self) -> str:
        """ A string that identifies this likelihood function """

    @property 
    @abstractmethod 
    def shape(self) -> Tuple[int]:
        """ Shape of the parameter grid on which this likelihood function
        is evaluated. """

    @property 
    @abstractmethod
    def parameter_names(self) -> Tuple[str]:
        """ Names of the parameters defining each axis of the parameter 
        grid """

    @property
    @abstractmethod
    def parameter_values(self) -> Tuple[np.ndarray]:
        """ Values of the parameters on which this likelihood function is
        evaluated. Returns a tuple of 1D np.ndarray defining the values
        along each parameter axis """

    @property 
    @abstractmethod 
    def parameter_units(self) -> Tuple[str]:
        """ Units in which each parameter axis is defined """

    @abstractmethod 
    def __call__(self, trajectories: TrajectoryGroup) -> Tuple[np.ndarray]:
        """ Evaluate the log likelihood function on each trajectory at each 
        point on the parameter grid. Returns two numpy.ndarrays encoding the 
        evaluated log likelihood function as the number of jumps per trajectory.
        The first numpy.ndarray has shape (*self.shape, trajectories.n_tracks) 
        and the second has shape (trajectories.n_tracks). """

    @abstractmethod
    def exp(self, log_L: np.ndarray) -> np.ndarray:
        """ Take the exponent of a log likelihood function in a numerically
        stable way appropriate for this likelihood function """

    @abstractmethod 
    def correct_for_defocalization(self, occs: np.ndarray, normalize: bool) -> np.ndarray:
        """ Correct a set of state occupations on this parameter grid for the 
        effects of defocalization. *occs* and the output of the function are 
        numpy.ndarrays of shape *self.shape*. """

    @abstractmethod
    def marginalize_on_diff_coef(self, occs: np.ndarray) -> np.ndarray:
        """ Marginalize a set of state occupations on the diffusion coefficient.
        May raise NotImplementedError if diffusion coefficient is not a parameter
        for the likelihood function (although this is the case for all 
        likelihoods implemented to date). 

        *occs* is a numpy.ndarray of shape *self.shape* and the return value
        is a 1D numpy.ndarray over the diffusion coefficient axis. """


class RBMELikelihood(Likelihood):
    """ Likelihood function for a regular Brownian motion with localization 
    error (RBME). This likelihood function has two parameters:
        1. the diffusion coefficient (specified in squared microns per sec)
        2. the localization error (specified in microns)

    The vector of all jumps along one axis for an RBME is a multivariate
    normal random vector with mean zero and covariance matrix *C* such that

        C[i,j] = 2 * (D * dt + loc_error^2)   if i == j
        C[i,j] = -loc_error^2                 if abs(i-j) == 1
        C[i,j] = 0                            otherwise

    where *dt* is the frame interval in seconds.

    The parameter grid for the RBMELikelihood object is a set of diffusion
    coefficients and localization errors. Localization error is usually 
    treated as a nuisance parameter (i.e. we marginalize over it after 
    inference).

    init
    ----
        pixel_size_um   :   size of camera pixels in microns
        frame_interval  :   frame interval in seconds
        focal_depth     :   microscope focal depth in microns
        diff_coefs      :   1D numpy.ndarray, the set of diffusion 
                            coefficients to use for the parameter grid
        loc_errors      :   1D numpy.ndarray, the set of localization 
                            errors to use for the parameter grid
    """
    def __init__(self, pixel_size_um: float, frame_interval: float,
        focal_depth: float=np.inf, diff_coefs: np.ndarray=DEFAULT_DIFF_COEFS,
        loc_errors: np.ndarray=DEFAULT_LOC_ERRORS, **kwargs):
        self.pixel_size_um = pixel_size_um 
        self.frame_interval = frame_interval
        self.focal_depth = focal_depth
        self.diff_coefs = np.asarray(diff_coefs)
        self.loc_errors = np.asarray(loc_errors)

    @property
    def name(self) -> str:
        return RBME
    
    @property 
    def shape(self) -> Tuple[int]:
        if not hasattr(self, "_shape"):
            self._shape = (self.diff_coefs.shape[0], self.loc_errors.shape[0])
        return self._shape 

    @property
    def parameter_names(self) -> Tuple[str]:
        if not hasattr(self, "_parameter_names"):
            self._parameter_names = ("diff_coef", "loc_error")
        return self._parameter_names

    @property 
    def parameter_values(self) -> Tuple[np.ndarray]:
        return (self.diff_coefs, self.loc_errors)

    @property 
    def parameter_units(self) -> Tuple[str]:
        if not hasattr(self, "_parameter_units"):
            self._parameter_units = ("µm2/sec", "µm")
        return self._parameter_units

    def __call__(self, trajectories: TrajectoryGroup) -> Tuple[np.ndarray]:
        """ Evaluate the log likelihood function specified by this *Likelihood*
        object on a set of target trajectories. 

        args
        ----
            trajectories    :   saspt.TrajectoryGroup, a set of target 
                                trajectories

        returns
        -------
            (
                3D numpy.ndarray with shape (*self.shape, n_tracks), the log 
                    likelihood function for each point on the parameter grid 
                    evaluated on each of the trajectories;

                1D numpy.ndarray of shape (n_tracks), the number of jumps in
                    each trajectory
            )
        """
        J = trajectories.jumps 
        n_tracks = trajectories.n_tracks

        # Generate the RBME jump covariance matrix
        def make_cov(diff_coef, loc_error, n) -> np.ndarray:
            le2 = loc_error ** 2
            C = 2 * (diff_coef * self.frame_interval + le2) * np.identity(n)
            for i in range(n-1):
                C[i,i+1] = -le2 
                C[i+1,i] = -le2 
            return C

        # Output
        log_L = np.empty((*self.shape, n_tracks), dtype=np.float64)
        jumps_per_track = np.empty(n_tracks, dtype=np.float64)

        # Iterate through the different trajectory lengths
        for n in range(1, trajectories.splitsize+1):

            # Get all track vectors matching this length. *V* is a 3D numpy.ndarray
            # with shape (M, 2, n), where *M* is the number of trajectories with 
            # exactly *n* jumps
            V, track_idx = trajectories.get_track_vectors(n)
            jumps_per_track[track_idx] = n
            y_jumps = V[:,0,:]
            x_jumps = V[:,1,:]

            for i, D in enumerate(self.diff_coefs):
                for j, le in enumerate(self.loc_errors):
                    C = make_cov(D, le, n)
                    C_inv = np.linalg.inv(C)
                    log_norm = n * np.log(2*np.pi) + np.linalg.slogdet(C)[1]
                    y_ll = ((y_jumps @ C_inv) * y_jumps).sum(axis=1)
                    x_ll = ((x_jumps @ C_inv) * x_jumps).sum(axis=1)
                    log_L[i,j,track_idx] = -0.5 * (y_ll + x_ll) - log_norm

        return log_L, jumps_per_track

    def exp(self, log_L: np.ndarray) -> np.ndarray:
        """ Given a log likelihood function, take the exponent to get the
        likelihood function and normalize over all states for each trajectory.

        args
        ----
            log_L   :   3D numpy.ndarray of shape (*self.shape, n_tracks),
                        log likelihoods of each trajectory at each point
                        on the parameter grid

        returns
        -------
            3D numpy.ndarray of shape (*self.shape, n_tracks),
                likelihoods of each track at each point on the
                parameter grid
        """
        if log_L.size > 0:
            log_L[np.isnan(log_L)] = -np.inf
            log_L[log_L>=100.0] = 100.0           
            L = np.exp(log_L - np.nanmax(log_L, axis=(0, 1)))
            return L / L.sum(axis=(0,1))
        else:
            return log_L.copy()

    def correct_for_defocalization(self, occs: np.ndarray, normalize: bool=False) -> np.ndarray:
        """ Apply a defocalization correction suitable for RBMEs to a set of state
        occupations.

        args
        ----
            occs        :   2D numpy.ndarray, state occupations with axes corresponding
                            to self.parameter_values
            normalize   :   bool, normalize over state occupations before returning

        returns
        -------
            2D numpy.ndarray of shape *occs*, the corrected state occupations
        """
        return defoc_corr(occs, self.parameter_values, likelihood=self.name, 
            frame_interval=self.frame_interval, dz=self.focal_depth, normalize=normalize)

    def marginalize_on_diff_coef(self, occs: np.ndarray) -> np.ndarray:
        """ Marginalize a set of state occupations or trajectory-state 
        assignments on the diffusion coefficients.

        args
        ----
            occs    :   numpy.ndarray of shape *self.shape*, state occupations,
                        or numpy.ndarray of shape (*self.shape, n_tracks), 
                        trajectory-state assignments

        returns
        -------
            numpy.ndarray, the state occupations or trajectory-state 
                assignments with parameters other than diffusion coefficient
                marginalized out
        """
        if len(occs.shape) == 2:
            out = occs.sum(axis=1)
            S = out.sum()
            return out / S if S > 0 else out 
        elif len(occs.shape) == 3:
            # Marginalize out localization error
            out = occs.sum(axis=1)
            # Normalization constant for each trajectory
            S = out.sum(axis=0)
            nonzero = S > 0
            out[:,nonzero] = out[:,nonzero] / S[nonzero]
            return out


class GammaLikelihood(Likelihood):
    def __init__(self, pixel_size_um: float, frame_interval: float,
        focal_depth: float, diff_coefs: np.ndarray=DEFAULT_DIFF_COEFS, 
        loc_error: float=0.035, mode: str="point", **kwargs):

        self.pixel_size_um = pixel_size_um
        self.frame_interval = frame_interval
        self.focal_depth = focal_depth  
        self.loc_error = loc_error
        self.diff_coefs = diff_coefs
        self.mode = mode

    @property
    def name(self) -> str:
        return GAMMA

    @property
    def shape(self) -> Tuple[int]:
        if not hasattr(self, "_shape"):
            self._shape = (self.diff_coefs.shape[0],)
        return self._shape

    @property
    def parameter_names(self) -> Tuple[str]:
        if not hasattr(self, "_parameter_names"):
            self._parameter_names = ("diff_coef",)
        return self._parameter_names 

    @property
    def parameter_values(self) -> Tuple[np.ndarray]:
        return (self.diff_coefs,)

    @property
    def parameter_units(self) -> Tuple[str]:
        if not hasattr(self, "_parameter_units"):
            self._parameter_units = ("µm2/sec",)
        return self._parameter_units

    def __call__(self, trajectories: TrajectoryGroup) -> Tuple[np.ndarray]:
        """ Evaluate the log likelihood for each trajectory at each point
        on the parameter grid.

        args
        ----
            trajectories    :   trajectories on which to evaluate the function

        returns
        -------
            (
                2D numpy.ndarray of shape (n_diff_coefs, n_tracks), the log
                    likelihood function;

                1D numpy.ndarray of shape (n_tracks,), the number of jumps
                    in each trajectory
            )
        """
        le2 = self.loc_error ** 2
        K = self.diff_coefs.shape[0]
        n_tracks = trajectories.n_tracks
        alpha = trajectories.jumps_per_track.astype(np.float64)
        sum_r2 = np.array(trajectories.jumps.groupby(TRACK)[DR2].sum())
        log_L = np.zeros((K, n_tracks), dtype=np.float64)
        for j in range(K):
            phi = 4 * (self.diff_coefs[j] * self.frame_interval + le2)
            log_L[j,:] = -(sum_r2 / phi) - alpha * np.log(phi)
        return log_L, alpha

    def exp(self, log_L: np.ndarray) -> np.ndarray:
        """ Given an evaluated log likelihood function, take the exponent
        and normalize to get the likelihood function.

        args
        ----
            log_L       :   2D numpy.ndarray of shape (n_diff_coefs, n_tracks),
                            the result of GammaLikelihood.__call__

        returns
        -------
            2D numpy.ndarray of shape *log_L.shape*, the corresponding
                normalized likelihood function
        """
        if log_L.size > 0:
            L = np.exp(log_L - log_L.max(axis=0))
            L = L / L.sum(axis=0)
            return L
        else:
            return log_L.copy()

    def correct_for_defocalization(self, occs: np.ndarray, normalize: bool=False) -> np.ndarray:
        return defoc_corr(L=occs, support=self.parameter_values, likelihood=self.name,
            frame_interval=self.frame_interval, dz=self.focal_depth, normalize=normalize)

    def marginalize_on_diff_coef(self, occs: np.ndarray) -> np.ndarray:
        """ Marginalize a set of state occupations or trajectory-state 
        assignments on the diffusion coefficients.

        args
        ----
            occs    :   numpy.ndarray of shape *self.shape*, state occupations,
                        or numpy.ndarray of shape (*self.shape, n_tracks), 
                        trajectory-state assignments

        returns
        -------
            numpy.ndarray, the state occupations or trajectory-state 
                assignments with parameters other than diffusion coefficient
                marginalized out
        """
        return occs.copy()


class RBMEMarginalLikelihood(Likelihood):
    def __init__(self, pixel_size_um: float, frame_interval: float, focal_depth: float=np.inf,
        diff_coefs: np.ndarray=DEFAULT_DIFF_COEFS, loc_errors: np.ndarray=DEFAULT_LOC_ERRORS,
        **kwargs):

        self.pixel_size_um = pixel_size_um 
        self.frame_interval = frame_interval
        self.focal_depth = focal_depth
        self.diff_coefs = diff_coefs
        self.loc_errors = loc_errors 

    @property
    def name(self) -> str:
        return RBME_MARGINAL

    @property
    def shape(self) -> Tuple[int]:
        if not hasattr(self, "_shape"):
            self._shape = (self.diff_coefs.shape[0],)
        return self._shape

    @property
    def parameter_names(self) -> Tuple[str]:
        if not hasattr(self, "_parameter_names"):
            self._parameter_names = ("diff_coef",)
        return self._parameter_names 

    @property
    def parameter_values(self) -> Tuple[np.ndarray]:
        return (self.diff_coefs,)

    @property
    def parameter_units(self) -> Tuple[str]:
        if not hasattr(self, "_parameter_units"):
            self._parameter_units = ("µm2/sec",)
        return self._parameter_units

    def __call__(self, trajectories: TrajectoryGroup) -> Tuple[np.ndarray]:
        J = trajectories.jumps 
        n_tracks = trajectories.n_tracks

        # Generate the RBME jump covariance matrix
        def make_cov(diff_coef, loc_error, n) -> np.ndarray:
            le2 = loc_error ** 2
            C = 2 * (diff_coef * self.frame_interval + le2) * np.identity(n)
            for i in range(n-1):
                C[i,i+1] = -le2 
                C[i+1,i] = -le2 
            return C

        # Output
        log_L = np.empty((self.diff_coefs.shape[0], self.loc_errors.shape[0], n_tracks), dtype=np.float64)
        jumps_per_track = np.empty(n_tracks, dtype=np.float64)

        # Iterate through the different trajectory lengths
        for n in range(1, trajectories.splitsize+1):

            # Get all track vectors matching this length. *V* is a 3D numpy.ndarray
            # with shape (M, 2, n), where *M* is the number of trajectories with 
            # exactly *n* jumps
            V, track_idx = trajectories.get_track_vectors(n)
            jumps_per_track[track_idx] = n
            y_jumps = V[:,0,:]
            x_jumps = V[:,1,:]

            for i, D in enumerate(self.diff_coefs):
                for j, le in enumerate(self.loc_errors):
                    C = make_cov(D, le, n)
                    C_inv = np.linalg.inv(C)
                    log_norm = n * np.log(2*np.pi) + np.linalg.slogdet(C)[1]
                    y_ll = ((y_jumps @ C_inv) * y_jumps).sum(axis=1)
                    x_ll = ((x_jumps @ C_inv) * x_jumps).sum(axis=1)
                    log_L[i,j,track_idx] = -0.5 * (y_ll + x_ll) - log_norm

        # Marginalize on localization error, letting log(0) = -inf
        with warnings.catch_warnings():
            warnings.simplefilter('ignore')
            log_L = np.log(self.exp(log_L).sum(axis=1))

        return log_L, jumps_per_track

    def exp(self, log_L: np.ndarray) -> np.ndarray:
        """ Given an evaluated log likelihood function, take the exponent
        and normalize to get the likelihood function.

        args
        ----
            log_L       :   2D numpy.ndarray of shape (n_diff_coefs, n_tracks),
                            the result of RBMEMarginalLikelihood.__call__

        returns
        -------
            2D numpy.ndarray of shape *log_L.shape*, the corresponding
                normalized likelihood function
        """
        if log_L.size > 0:
            L = np.exp(log_L - log_L.max(axis=0))
            L = L / L.sum(axis=0)
            return L
        else:
            return log_L.copy()

    def correct_for_defocalization(self, occs: np.ndarray, normalize: bool=False) -> np.ndarray:
        return defoc_corr(occs, self.parameter_values, likelihood=self.name,
            frame_interval=self.frame_interval, dz=self.focal_depth, normalize=normalize)

    def marginalize_on_diff_coef(self, occs: np.ndarray) -> np.ndarray:
        """ Marginalize a set of state occupations or trajectory-state 
        assignments on the diffusion coefficients.

        args
        ----
            occs    :   numpy.ndarray of shape *self.shape*, state occupations,
                        or numpy.ndarray of shape (*self.shape, n_tracks), 
                        trajectory-state assignments

        returns
        -------
            numpy.ndarray, the state occupations or trajectory-state 
                assignments with parameters other than diffusion coefficient
                marginalized out
        """
        return occs.copy()


class FBMELikelihood(Likelihood):
    def __init__(
        self,
        pixel_size_um: float,
        frame_interval: float, 
        focal_depth: float,
        diff_coefs: np.ndarray=DEFAULT_DIFF_COEFS,
        hurst_pars: np.ndarray=DEFAULT_HURST_PARS,
        loc_errors: np.ndarray=DEFAULT_LOC_ERRORS,
        **kwargs
    ):
        self.pixel_size_um = pixel_size_um 
        self.frame_interval = frame_interval
        self.focal_depth = focal_depth 
        self.diff_coefs = np.asarray(diff_coefs)
        self.hurst_pars = np.asarray(hurst_pars)
        self.loc_errors = np.asarray(loc_errors)

    @property
    def name(self) -> str:
        return FBME
    
    @property 
    def shape(self) -> Tuple[int]:
        if not hasattr(self, "_shape"):
            self._shape = (self.diff_coefs.shape[0], self.hurst_pars.shape[0], self.loc_errors.shape[0])
        return self._shape 

    @property
    def parameter_names(self) -> Tuple[str]:
        if not hasattr(self, "_parameter_names"):
            self._parameter_names = ("diff_coef", "hurst_parameter", "loc_error")
        return self._parameter_names

    @property 
    def parameter_values(self) -> Tuple[np.ndarray]:
        return self.diff_coefs, self.hurst_pars, self.loc_errors

    @property 
    def parameter_units(self) -> Tuple[str]:
        if not hasattr(self, "_parameter_units"):
            self._parameter_units = ("µm2/sec", "", "µm")
        return self._parameter_units

    def __call__(self, trajectories: TrajectoryGroup) -> Tuple[np.ndarray]:
        J = trajectories.jumps
        n_tracks = trajectories.n_tracks

        # Generate the FBME jump covariance matrix
        def make_cov(diff_coef, hurst_par, le, n) -> np.ndarray:
            le2 = le ** 2
            h2 = hurst_par * 2
            T, S = np.indices((n, n)) + 1
            C = diff_coef * self.frame_interval * (
                np.power(np.abs(T - S + 1), h2) + 
                np.power(np.abs(T - S - 1), h2) - 
                2 * np.power(np.abs(T - S), h2)
            )
            C += (2 * le2 * np.identity(n))
            for i in range(n-1):
                C[i,i+1] -= le2 
                C[i+1,i] -= le2 
            return C 

        # Output
        log_L = np.empty((*self.shape, n_tracks), dtype=np.float64)
        jumps_per_track = np.empty(n_tracks, dtype=np.float64)

        # Iterate through the different trajectory lengths
        for n in range(1, trajectories.splitsize+1):
            # Get all track vectors matching this length. *V* is a 3D numpy.ndarray
            # with shape (M, 2, n), where *M* is the number of trajectories with 
            # exactly *n* jumps
            V, track_idx = trajectories.get_track_vectors(n)
            jumps_per_track[track_idx] = n
            y_jumps = V[:,0,:]
            x_jumps = V[:,1,:]

            for i, D in enumerate(self.diff_coefs):
                for j, hp in enumerate(self.hurst_pars):
                    for k, le in enumerate(self.loc_errors):
                        C = make_cov(D, hp, le, n)
                        C_inv = np.linalg.inv(C)
                        log_norm = n * np.log(2*np.pi) + np.linalg.slogdet(C)[1]
                        y_ll = ((y_jumps @ C_inv) * y_jumps).sum(axis=1)
                        x_ll = ((x_jumps @ C_inv) * x_jumps).sum(axis=1)
                        log_L[i,j,k,track_idx] = -0.5 * (y_ll + x_ll) - log_norm

        return log_L, jumps_per_track

    def exp(self, log_L: np.ndarray) -> np.ndarray:
        """ Given a log likelihood function, take the exponent to get the
        likelihood function and normalize over all states for each trajectory.

        args
        ----
            log_L   :   4D numpy.ndarray of shape (*self.shape, n_tracks),
                        log likelihoods of each trajectory at each point
                        on the parameter grid

        returns
        -------
            4D numpy.ndarray of shape (*self.shape, n_tracks),
                likelihoods of each track at each point on the
                parameter grid
        """
        if log_L.size > 0:
            log_L[np.isnan(log_L)] = -np.inf
            log_L[log_L>=100.0] = 100.0
            L = np.exp(log_L - np.nanmax(log_L, axis=(0,1,2)))
            return L / L.sum(axis=(0,1,2))
        else:
            return log_L.copy()

    def correct_for_defocalization(self, occs: np.ndarray, normalize: bool=False) -> np.ndarray:
        if np.isinf(self.focal_depth):
            return occs
        else:
            return defoc_corr(occs, self.parameter_values, likelihood=self.name,
                frame_interval=self.frame_interval, dz=self.focal_depth, normalize=normalize)

    def marginalize_on_diff_coef(self, occs: np.ndarray) -> np.ndarray:
        """ Marginalize a set of state occupations or trajectory-state 
        assignments on the diffusion coefficients.

        args
        ----
            occs    :   numpy.ndarray of shape *self.shape*, state occupations,
                        or numpy.ndarray of shape (*self.shape, n_tracks), 
                        trajectory-state assignments

        returns
        -------
            numpy.ndarray, the state occupations or trajectory-state 
                assignments with parameters other than diffusion coefficient
                marginalized out
        """
        if len(occs.shape) == 3:
            out = occs.sum(axis=(1,2))
            S = out.sum()
            return out / S if S > 0 else out 
        elif len(occs.shape) == 4:
            # Marginalize out Hurst parameter and localization error magnitude
            out = occs.sum(axis=(1,2))
            # Normalization constant for each trajectory
            S = out.sum(axis=0)
            nonzero = S > 0
            out[:,nonzero] = out[:,nonzero] / S[nonzero]
            return out


####################################
## AVAILABLE LIKELIHOOD FUNCTIONS ##
####################################


LIKELIHOODS = {
    RBME: RBMELikelihood,
    GAMMA: GammaLikelihood,
    RBME_MARGINAL: RBMEMarginalLikelihood,
    FBME: FBMELikelihood,
}


def make_likelihood(likelihood_type: str, **kwargs):
    """ Factory method for Likelihood subclass instances. 

    args
    ----
        likelihood_type     :   a key in *LIKELIHOODS*

    returns
    -------
        new instance of the corresponding Likelihood subclass
    """
    if likelihood_type not in LIKELIHOODS.keys():
        avail = ", ".join(list(LIKELIHOODS.keys()))
        raise KeyError(f"likelihood {likelihood_type} not recognized; options are: {avail}")
    return LIKELIHOODS[likelihood_type](**kwargs)

def make_likelihood_from_params(likelihood_type: str, params: StateArrayParameters,
    **kwargs):
    """ Convenience function; wrapper on *make_likelihood* that extracts
    relevant parameters from a StateArrayParameters instance. 

    args
    ----
        likelihood_type     :   a key in *LIKELIHOODS*
        params              :   parameters for running a state array
        kwargs              :   any additional keyword arguments to the 
                                Likelihood constructor

    returns
    -------
        new instance of the corresponding Likelihood subclass 
    """
    kwargs = {
        'pixel_size_um': params.pixel_size_um,
        'frame_interval': params.frame_interval,
        'focal_depth': params.focal_depth, 
        **kwargs
    }
    return make_likelihood(likelihood_type, **kwargs)



import os, numpy as np, pandas as pd 
from typing import List
from functools import reduce 

from .io import load_detections
from .utils import track_length
from .constants import (
    TRACK, FRAME, PY, PX, TRACK_LENGTH,
    DFRAMES, DY, DX, DR2, JUMPS_PER_TRACK,
    DEFAULT_MAX_ITER,
    DEFAULT_SPLITSIZE,
    DEFAULT_CONC_PARAM,
    DEFAULT_START_FRAME,
)
from .parameters import StateArrayParameters

class TrajectoryGroup:
    """ Encapsulates a set of trajectories to be analyzed by a StateArray,
    caching attributes required for downstream steps.

    init
    ----
        detections      :   pandas.DataFrame indexed by detection with at
                            minimum the columns TRACK, FRAME, PY, and PX,
                            representing a set of trajectories

        pixel_size_um   :   size of pixels in microns

        frame_interval  :   time between camera frames in seconds

        splitsize       :   maximum number of jumps per trajectory. Trajectories
                            longer than this are split into smaller pieces.

    """
    # Column in TrajectoryGroup.detections encoding the original trajectory
    # index before preprocessing
    ORIG_TRACK = "orig_trajectory"

    # Column in TrajectoryGroup.detections encoding the original detection
    # index before preprocessing
    DETECT_INDEX = "detect_index"

    # Names of statistics in TrajectoryGroup.raw_track_statistics and 
    # TrajectoryGroup.processed_track_statistics
    statistic_names = ['n_tracks', 'n_jumps', 'n_detections',
        'mean_track_length', 'max_track_length', 'fraction_singlets',
        'fraction_unassigned', 'mean_jumps_per_track', 'mean_detections_per_frame',
        'max_detections_per_frame', 'fraction_of_frames_with_detections']

    def __init__(self, detections: pd.DataFrame, pixel_size_um: float,
        frame_interval: float, splitsize: int=DEFAULT_SPLITSIZE,
        start_frame: int=DEFAULT_START_FRAME, **kwargs):

        # Store a copy of the _unpreprocessed_ detections. This is useful for
        # calculating dataset statistics and localization density.
        self._raw_detections = detections[[TRACK, FRAME, PY, PX]].sort_values(
            by=[TRACK, FRAME]).reset_index(drop=True)
        self._raw_detections = track_length(self._raw_detections)

        # Preprocess trajectories for state array operations. This includes removing
        # singlets, breaking long trajectories into pieces of maximum size *splitsize*,
        # and reindexing for contiguous trajectory indices
        self.detections = self.preprocess(detections, splitsize, start_frame)

        # Some dataset parameters
        self.pixel_size_um = pixel_size_um
        self.frame_interval = frame_interval
        self.splitsize = splitsize
        self.start_frame = start_frame

        # We cannot work with input that contains multiple detections in the same
        # frame assigned to the same trajectory
        if len(self.detections) > 0:
            G = self.detections.loc[self.detections[TRACK] >= 0].groupby(FRAME)
            if (np.asarray(G[TRACK].nunique()) != np.asarray(G.size())).all():
                raise ValueError("*detections* cannot contain multiple detections in the " \
                    "same frame that are assigned to the same trajectory")

    def __enter__(self):
        return self

    def __exit__(self, etype, evalue, traceback):
        return etype is None

    def __repr__(self):
        return "TrajectoryGroup:\n{}".format("\n".join([
            f"{attr}:\t{getattr(self, attr)}" for attr in ["n_detections", "n_jumps", "n_tracks"]
        ]))

    @classmethod
    def from_params(cls, detections: pd.DataFrame, params: StateArrayParameters):
        """ Alternative constructor; instantiate a TrajectoryGroup using a
        StateArrayParameters object """
        kwargs = {k: getattr(params, k) for k in 
            ["pixel_size_um", "frame_interval", "splitsize", "start_frame"]}
        return cls(detections, **kwargs)

    @classmethod
    def from_files(cls, detection_files: List[str], **kwargs):
        """ Alternative constructor; instantiate a TrajectoryGroup directly
        from files containing detections """
        return cls(load_detections(*detection_files), **kwargs)

    @classmethod
    def preprocess(cls, detections: pd.DataFrame, splitsize: int=DEFAULT_SPLITSIZE,
        start_frame: int=DEFAULT_START_FRAME):
        """ Preprocess some trajectories for analysis with state arrays.

        This involves three steps:
            1. We remove all singlets, unassigned detections, and detections
                before *start_frame*.

            2. Long trajectories are split into pieces with a maximum
                size of *splitsize*.

            3. Trajectories are reindexed so that trajectory indices are 
                contiguous, starting from 0 to n_tracks-1.

        args
        ----
            detections      :   pandas.DataFrame indexed by detection. Must
                                contain the columns FRAME, TRACK, PY, and PX.

            splitsize       :   maximum length of trajectories in jumps

            start_frame     :   disregard detections before this frame

        returns
        -------
            pandas.DataFrame, processed trajectories. The old trajectory
                index for each detection is recorded under the
                *TrajectoryGroup.ORIG_TRACK* column, while the new
                trajectory index is under the *TRACK* column.
        """
        # Assign each detection a unique index, to keep track of them after sorting
        detections[cls.DETECT_INDEX] = np.arange(len(detections), dtype=np.int64)

        # Discard singlets, unassigned detections, and detections before the start frame
        detections = track_length(detections)
        detections = detections[reduce(np.logical_and, [
            detections[TRACK] >= 0,
            detections[TRACK_LENGTH] > 1,
            detections[FRAME] >= start_frame
        ])].sort_values(by=[TRACK, FRAME]).reset_index(drop=True)

        # Break trajectories longer than *splitsize* into smaller pieces
        orig_indices = np.array(detections[TRACK])
        detections[cls.ORIG_TRACK] = orig_indices
        detections[TRACK] = cls.split_tracks(orig_indices, splitsize)

        # Drop singlets generated by the splitting process
        detections = track_length(detections)
        detections = detections[detections[TRACK_LENGTH] > 1].reset_index(drop=True)

        # Reindex so that trajectory index is contiguous
        M = {t: i for i, t in enumerate(detections[TRACK].unique())}
        detections[TRACK] = detections[TRACK].map(M)

        return detections

    @classmethod
    def split_tracks(cls, track_indices: np.ndarray, splitsize: int) -> np.ndarray:
        """ Split trajectories into smaller pieces that have a maximum of *splitsize*
        jumps. 

        args
        ----
            track_indices       :   1D np.ndarray with the current trajectory
                                    index of each detection
            splitsize           :   int, maximum number of jumps to tolerate

        returns
        -------
            1D np.ndarray, dtype int64, the new set of trajectory indices
        """
        if track_indices.shape[0] > 1 and \
            ((track_indices[1:]-track_indices[:-1])<0).any():
            raise ValueError("track_indices must be monotonically increasing")
        if track_indices.shape[0] == 0:
            return np.array([]).astype(np.int64)
        new_indices = np.zeros(len(track_indices), dtype=np.int64)
        c = 0
        L = 0
        prev_index = track_indices[0]
        for i, index in enumerate(track_indices):
            if index == prev_index and L > splitsize:
                L = 0
                c += 1
            elif index != prev_index:
                prev_index = index
                L = 0
                c += 1
            new_indices[i] = c
            L += 1
        return new_indices 

    ################
    ## PROPERTIES ##
    ################

    @property 
    def n_detections(self) -> int:
        if not hasattr(self, "_n_detections"):
            self._n_detections = len(self.detections)
        return self._n_detections

    @property 
    def n_jumps(self) -> int:
        if not hasattr(self, "_n_jumps"):
            self._n_jumps = len(self.jumps)
        return self._n_jumps

    @property 
    def n_tracks(self) -> int:
        if not hasattr(self, "_n_tracks"):
            self._n_tracks = self.detections[TRACK].nunique()
        return self._n_tracks

    @property 
    def jumps(self) -> pd.DataFrame:
        if not hasattr(self, "_jumps"):
            T = np.asarray(self.detections[[FRAME, FRAME, TRACK, PY, PX, PX]])
            T[:,3:5] *= self.pixel_size_um
            if T.shape[0] < 2:
                self._jumps = pd.DataFrame({FRAME: np.zeros(0, dtype=np.int64),
                    DFRAMES: np.zeros(0, dtype=np.int64), TRACK: np.zeros(0, dtype=np.int64),
                    DY: np.zeros(0, dtype=np.float64), DX: np.zeros(0, dtype=np.float64),
                    DR2: np.zeros(0, dtype=np.float64), JUMPS_PER_TRACK: np.zeros(0, dtype=np.int64)})
            else:
                # Diff
                D = T[1:,:] - T[:-1,:]
                # Map back trajectory and frame indices
                D[:,0] = T[:-1,0]
                # Detections that originate from the same trajectory
                same_track = D[:,2] == 0
                D[:,2] = T[:-1,2]
                D = D[same_track,:]
                # Calculate 2D radial jumps
                D[:,5] = (D[:,3:5]**2).sum(axis=1) / D[:,1]
                # Format as pandas.DataFrame
                self._jumps = pd.DataFrame(D, columns=[FRAME, DFRAMES, TRACK, DY, DX, DR2])
                # Enforce some types
                for c in [FRAME, DFRAMES, TRACK]:
                    self._jumps[c] = self._jumps[c].astype(np.int64)
                # Calculate number of jumps per track
                self._jumps = self._jumps.join(
                    self._jumps.groupby(TRACK).size().rename(JUMPS_PER_TRACK),
                    on=TRACK)
        return self._jumps

    @property
    def jumps_per_track(self) -> np.ndarray:
        """ The number of jumps per trajectory.

        returns
        -------
            1D numpy.ndarray of shape (n_tracks,), the number of jumps in each
                trajectory
        """
        if not hasattr(self, "_jumps_per_track"):
            self._jumps_per_track = np.array(self.jumps.groupby(TRACK)[JUMPS_PER_TRACK].first())
        return self._jumps_per_track

    @property 
    def raw_track_statistics(self) -> dict:
        """ Return some statistics on the raw unprocessed trajectories that 
        are wrapped by this TrajectoryGroup object.

        returns
        -------
            dict keyed by statistic name
        """
        if not hasattr(self, "_raw_track_statistics"):
            # Assigned trajectories
            T = self._raw_detections.loc[self._raw_detections[TRACK] >= 0]
            n_frames = 1 + self._raw_detections[FRAME].max() - self._raw_detections[FRAME].min()
            self._raw_track_statistics = {
                "n_tracks": T[TRACK].nunique(),
                "n_jumps": (T.groupby(TRACK).size() - 1).sum(),
                "n_detections": len(self._raw_detections),
                "mean_track_length": T.groupby(TRACK)[TRACK_LENGTH].first().mean(),
                "max_track_length": T.groupby(TRACK)[TRACK_LENGTH].first().max() \
                    if len(T)>0 else 0,
                "fraction_singlets": (T.groupby(TRACK)[TRACK_LENGTH].first() \
                    == 1).mean(),
                "fraction_unassigned": (self._raw_detections[TRACK] < 0).mean(),
                "mean_jumps_per_track": (T.groupby(TRACK).size() - 1).mean(),
                "mean_detections_per_frame": len(self._raw_detections) / n_frames,
                "max_detections_per_frame": self._raw_detections.groupby(FRAME).size().max() \
                    if len(self._raw_detections)>0 else 0,
                "fraction_of_frames_with_detections": self._raw_detections[FRAME].nunique() \
                    / n_frames if len(self._raw_detections)>0 else 0.0,
            }
        return self._raw_track_statistics

    @property 
    def processed_track_statistics(self) -> dict:
        """ Return some statistics on the preprocessed trajectories (i.e. after
        calling self.preprocess).

        self.preprocess does three things:
            - Removes singlets and detections unassigned to any trajectory
            - Splits big trajectories into smaller pieces (max size *splitsize*)
            - Reindexes trajectories for contiguous trajectory indices

        returns
        -------
            dict keyed by statistic name
        """
        if not hasattr(self, "_processed_track_statistics"):
            n_frames = 1 + self.detections[FRAME].max() - self.start_frame
            self._processed_track_statistics = {
                "n_tracks": self.n_tracks,
                "n_jumps": self.n_jumps,
                "n_detections": self.n_detections,
                "mean_track_length": self.detections.groupby(TRACK)[TRACK_LENGTH].first().mean(),
                "max_track_length": self.detections.groupby(TRACK)[TRACK_LENGTH].first().max() \
                    if len(self.detections)>0 else 0,
                "fraction_singlets": (self.detections.groupby(TRACK)[TRACK_LENGTH].first() \
                    == 1).mean(),
                "fraction_unassigned": (self.detections[TRACK] < 0).mean(),
                "mean_jumps_per_track": self.jumps.groupby(TRACK).size().mean(),
                "mean_detections_per_frame": len(self.detections) / n_frames,
                "max_detections_per_frame": self.detections.groupby(FRAME).size().max() \
                    if len(self.detections)>0 else 0,
                "fraction_of_frames_with_detections": self.detections[FRAME].nunique() \
                    / n_frames if len(self.detections)>0 else 0.0,
            }
        return self._processed_track_statistics

    #############
    ## METHODS ##
    #############

    def get_track_vectors(self, n: int) -> np.ndarray:
        """ Return the vectorial jumps corresponding to all trajectories with 
        exactly *n* jumps.

        args
        ----
            n       :   number of jumps per track

        returns
        -------
            (
                3D np.ndarray of shape (M, 2, n), the Y and X jumps
                    of each trajectory (where *M* is the number of trajectories
                    that have *n* jumps);

                1D np.ndarray of shape (M,), the trajectory indices of each
                    trajectory
            )
        """
        take = self.jumps[JUMPS_PER_TRACK] == n
        track_indices = np.array(self.jumps.loc[take, TRACK].unique())
        M = take.sum() // n if n>0 else 0
        V = np.empty((M, 2, n), dtype=np.float64)
        V[:,0,:] = self.jumps.loc[take, DY].to_numpy(copy=True).reshape((M, n))
        V[:,1,:] = self.jumps.loc[take, DX].to_numpy(copy=True).reshape((M, n))
        return V, track_indices

    def subsample(self, size: int):
        """ Generate another TrajectoryGroup object by subsampling the trajectories
        in the present object without replacement.

        args
        ----
            size        :   number of trajectories to sample

        returns
        -------
            new instance of TrajectryGroup
        """
        if size > self.n_tracks:
            size = self.n_tracks
        detections = self.detections[self.detections[TRACK].isin(
            np.random.choice(np.arange(self.n_tracks), size=size, replace=False)
        )].reset_index(drop=True)
        return TrajectoryGroup(detections, pixel_size_um=self.pixel_size_um,
            frame_interval=self.frame_interval, splitsize=self.splitsize, 
            start_frame=self.start_frame)
